unit UMain;

{
Copyright (c) 2014, Fernando Biazi Nascimento
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies,
either expressed or implied, of any organization.
}

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, TAGraph, TADbSource, TASeries,
  TATransformations, Forms, Controls, Dialogs, Buttons, StdCtrls, ExtCtrls,
  ComCtrls, Menus, ActnList, StdActns, DBGrids, SdpoSerial, FBNBufDataset,
  dateutils, DB, LCLType, Clipbrd, maskedit, Spin, TACustomSource;

type

  { TDBGridSO }

  TDBGridSO = class(TDBGrid)
  public
    property SaveOptions;
  end;

  { TfrmMain }

  TfrmMain = class(TForm)
    actConnOpen:     TAction;
    actConnClose:    TAction;
    actEnumerateSerial: TAction;
    actFileExportSaveDialog: TSaveDialog;
    actFileNew:      TAction;
    actFileSave:     TAction;
    actFileClean:    TAction;
    actFileSaveAsSaveDialog: TSaveDialog;
    actMemoryDump:   TAction;
    actPacketsCopyLine: TAction;
    actViewConnPanel: TAction;
    actSeparator:    TAction;
    actSetClock:     TAction;
    actMakeSink:     TAction;
    actResetMSP430:  TAction;
    actResetCC2480:  TAction;
    actSetEndDevice: TAction;
    actSetRouter:    TAction;
    actSetCoord:     TAction;
    ActionList1:     TActionList;
    bdsCommands:     TFBNBufDataset;
    bdsCommandscmdId: TWordField;
    bdsCommandscmdName: TStringField;
    bdsPackets:      TFBNBufDataset;
    bdsPacketscmdHex: TStringField;
    bdsPacketscmdName: TStringField;
    bdsPacketsCommand: TWordField;
    bdsPacketsData:  TBytesField;
    bdsPacketsDataHex: TStringField;
    bdsPacketsId:    TAutoIncField;
    bdsPacketsLength: TSmallintField;
    bdsPacketsTimeStamp: TDateTimeField;
    bdsPacketsTimeStampMs: TWordField;
    bdsPacketsTimeStampUnix: TLargeintField;
    bdsPTPTiming:    TFBNBufDataset;
    bdsPTPTimingDelayToMaster: TFloatField;
    bdsPTPTimingDeviceHex: TStringField;
    bdsPTPTimingId:  TLongintField;
    bdsPTPTimingLastAdjust: TLargeintField;
    bdsPTPTimingPTPTime: TFloatField;
    bdsPTPTimingPTPTimeNS: TLongintField;
    bdsPTPTimingPTPTimeS: TLargeintField;
    bdsSensors:      TFBNBufDataset;
    bdsSensorsDeviceHex: TStringField;
    bdsSensorsId:    TLongintField;
    bdsSensorsLumTime: TFloatField;
    bdsSensorsLumTimeNS: TLongintField;
    bdsSensorsLumTimeS: TLargeintField;
    bdsSensorsLumValue: TFloatField;
    bdsSensorsTempTime: TFloatField;
    bdsSensorsTempTimeNS: TLongintField;
    bdsSensorsTempTimeS: TLargeintField;
    bdsSensorsTempValue: TFloatField;
    bdsSensorsVoltTime: TFloatField;
    bdsSensorsVoltTimeNS: TLongintField;
    bdsSensorsVoltTimeS: TLargeintField;
    bdsSensorsVoltValue: TFloatField;
    Bevel1:          TBevel;
    btnCloseSerial:  TButton;
    btnOpenSerial:   TButton;
    btnPacketDump:   TButton;
    btnPacketExport: TButton;
    btnPTPExport:    TButton;
    btnSensorExport: TButton;
    catLumAs:        TAutoScaleAxisTransform;
    catPTPTimingDelay: TChartAxisTransformations;
    catPTPTimingAdjust: TChartAxisTransformations;
    catPTPTimingAdjustAs: TAutoScaleAxisTransform;
    catTempAs:       TAutoScaleAxisTransform;
    catPTPTimingDelayAs: TAutoScaleAxisTransform;
    catVoltAs:       TAutoScaleAxisTransform;
    cbRollPacketList: TCheckBox;
    catTemp:         TChartAxisTransformations;
    catVolt:         TChartAxisTransformations;
    catLum:          TChartAxisTransformations;
    cbRollSensorList: TCheckBox;
    cbRollPtpList:   TCheckBox;
    chtPTPTiming:    TChart;
    chtPTPTimingLsAdjust: TLineSeries;
    chtPTPTimingLsDelay: TLineSeries;
    chtSensors:      TChart;
    chtSensorsLsVolt: TLineSeries;
    chtSensorsLsLum: TLineSeries;
    chtSensorsLsTemp: TLineSeries;
    cmbBaudRate:     TComboBox;
    cmbDataBits:     TComboBox;
    cmbParity:       TComboBox;
    cmbPTPTimingDevice: TComboBox;
    cmbSerialSelect: TComboBox;
    cmbStopBits:     TComboBox;
    actFileExit:     TFileExit;
    cmbSensorsDevice: TComboBox;
    dsSensors:       TDatasource;
    dbgSensors:      TDBGrid;
    dcsPTPTimingDelay: TDbChartSource;
    dcsPTPTimingAdjust: TDbChartSource;
    dsPTPTiming:     TDatasource;
    dcsSensorsTemp:  TDbChartSource;
    dcsSensorsVolt:  TDbChartSource;
    dcsSensorsLum:   TDbChartSource;
    dbgPTPTiming:    TDBGrid;
    dsPackets:       TDatasource;
    dbgPackets:      TDBGridSO;
    actFileOpen:     TFileOpen;
    actFileSaveAs:   TFileSaveAs;
    actFileExport:   TFileSaveAs;
    actHelpLicense:  THelpAction;
    imlActions:      TImageList;
    Label1:          TLabel;
    Label3:          TLabel;
    MainMenu1:       TMainMenu;
    miLicense:       TMenuItem;
    miFileExport:    TMenuItem;
    miDividerFile4:  TMenuItem;
    miDividerFile2:  TMenuItem;
    miFileClean:     TMenuItem;
    miDividerFile1:  TMenuItem;
    mkeDumpAddress:  TMaskEdit;
    pnlPtpData:      TPanel;
    pnlSensorListOptions: TPanel;
    pnlSensorData:   TPanel;
    pnlPtpListOptions1: TPanel;
    ppmPacketsGridmiCopyLine: TMenuItem;
    miConnSetup:     TMenuItem;
    miView:          TMenuItem;
    miDividerFile3:  TMenuItem;
    miFileSaveAs:    TMenuItem;
    miFileSave:      TMenuItem;
    miFileOpen:      TMenuItem;
    miFileNew:       TMenuItem;
    miSeparator1:    TMenuItem;
    miEnumerateSerial: TMenuItem;
    miConnClose:     TMenuItem;
    miConnOpen:      TMenuItem;
    miConnection:    TMenuItem;
    miSetClock:      TMenuItem;
    miMakeSink:      TMenuItem;
    miResetMSP430:   TMenuItem;
    miResetCC2480:   TMenuItem;
    miSeparator2:    TMenuItem;
    miReset:         TMenuItem;
    miSetEndDevice:  TMenuItem;
    miSetRouter:     TMenuItem;
    miSetCoord:      TMenuItem;
    miDeviceSetType: TMenuItem;
    miDevice:        TMenuItem;
    miFileExit:      TMenuItem;
    miFile:          TMenuItem;
    mmoTextRcv:      TMemo;
    OpenDialog:      TOpenDialog;
    PageControl1:    TPageControl;
    pnlSensorsInfo:  TPanel;
    pnlPTPGraph:     TPanel;
    pnSensorGraph:   TPanel;
    pnlPTPTiming:    TPanel;
    pnlToolBars:     TPanel;
    pnlPacketListOptions: TPanel;
    pnlPacketList:   TPanel;
    pnlCtrl:         TPanel;
    pnlConnSetup:    TPanel;
    ppmPacketsGrid:  TPopupMenu;
    SaveDialog:      TSaveDialog;
    Serial:          TSdpoSerial;
    spbEnumerateSerial: TSpeedButton;
    speDumpSize:     TSpinEdit;
    Splitter1:       TSplitter;
    Splitter2:       TSplitter;
    Splitter3:       TSplitter;
    StatusBar1:      TStatusBar;
    tbtConnOpen:     TToolButton;
    tbtConnClose:    TToolButton;
    tbtDivider3:     TToolButton;
    tsSensors:       TTabSheet;
    tsPackets:       TTabSheet;
    tsPTP:           TTabSheet;
    tbrDevice:       TToolBar;
    tbtSetCoord:     TToolButton;
    tbtSetRouter:    TToolButton;
    tbtSetEndDevice: TToolButton;
    tbtDivider1:     TToolButton;
    tbtReset:        TToolButton;
    tbtResetCC2480:  TToolButton;
    tbtDivider2:     TToolButton;
    tbtSetNetTime:   TToolButton;
    tbtSink:         TToolButton;
    tbrView:         TToolBar;
    ToolConnSetup:   TToolButton;
    procedure actConnCloseExecute(Sender: TObject);
    procedure actConnOpenExecute(Sender: TObject);
    procedure actEnumerateSerialExecute(Sender: TObject);
    procedure actFileCleanExecute(Sender: TObject);
    procedure actFileExportAccept(Sender: TObject);
    procedure actFileNewExecute(Sender: TObject);
    procedure actFileOpenAccept(Sender: TObject);
    procedure actFileSaveAsAccept(Sender: TObject);
    procedure actFileSaveAsBeforeExecute(Sender: TObject);
    procedure actFileSaveExecute(Sender: TObject);
    procedure actMakeSinkExecute(Sender: TObject);
    procedure actMemoryDumpExecute(Sender: TObject);
    procedure actResetCC2480Execute(Sender: TObject);
    procedure actResetMSP430Execute(Sender: TObject);
    procedure actSetClockExecute(Sender: TObject);
    procedure actSetCoordExecute(Sender: TObject);
    procedure actSetEndDeviceExecute(Sender: TObject);
    procedure actSetRouterExecute(Sender: TObject);
    procedure actViewConnPanelExecute(Sender: TObject);
    procedure bdsPacketsAfterInsert(DataSet: TDataSet);
    procedure bdsPacketsAfterPost(DataSet: TDataSet);
    procedure bdsPacketsCalcFields(DataSet: TDataSet);
    procedure bdsPTPTimingCalcFields(DataSet: TDataSet);
    procedure bdsSensorsCalcFields(DataSet: TDataSet);
    procedure btnPTPExportClick(Sender: TObject);
    procedure btnSensorExportClick(Sender: TObject);
    procedure cmbPTPTimingDeviceChange(Sender: TObject);
    procedure cmbSensorsDeviceChange(Sender: TObject);
    procedure cmbSerialSelectChange(Sender: TObject);
    procedure dbgPacketsCellClick(Column: TColumn);
    procedure dbgPacketsKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure dbgPTPTimingKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure dbgSensorsKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure dcsSensorsTempGetItem(ASender: TDbChartSource;
      var AItem: TChartDataItem);
    procedure FormActivate(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure actPacketsCopyLineExecute(Sender: TObject);
    procedure actHelpLicenseExecute(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure mkeDumpAddressExit(Sender: TObject);
    procedure mkeDumpAddressKeyPress(Sender: TObject; var Key: Char);
    procedure SerialRxData(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
  end;

var
  frmMain: TfrmMain;

procedure copyPacketToGraphPTP(pDS, ptpDS: TDataSet; device: String);
procedure copyPacketToGraphSensors(pDS, sDS: TDataSet; device: String);
procedure copyPacketsToGraphPTP(pDS, ptpDS: TDataSet; cmb: TComboBox);
procedure copyPacketsToGraphSensors(pDS, sDS: TDataSet; cmb: TComboBox);
procedure FillmdsCommands(DataSet: TDataSet);
function getPTPClockAcuracySpecification(n: Byte): String;
function getPTPMessageTypeName(n: Byte): String;
function getPTPTimeSourceName(n: Byte): String;
procedure processPacket(var buf: AnsiString; DataSet: TDataSet; keepCurrent: Boolean);
function showInvalidLenPacket(expectedLen, Len: Byte; strHexBuf: AnsiString;
  Memo: TMemo): Boolean;
procedure showHeartBeat(pBuf: PByte; size: Byte; Memo: TMemo);
procedure showPacket(DataSet: TDataSet; Memo: TMemo);
procedure showPtpReport(pBuf: PByte; Memo: TMemo);
procedure showSensorReport(pBuf: PByte; size: Byte; Memo: TMemo);
procedure showRawData(header: String; var buf: AnsiString; Memo: TMemo);
procedure sendCommand(cmd: ShortString; serial: TSdpoSerial);
procedure sendTickSize(newTickSize: Cardinal);
procedure sendTimeAdjust(adjNs: Int64);
function strIif(test: Boolean; truePart, falsePart: String): String;

implementation

{$R *.lfm}

uses
  registry, versionresource, versiontypes, ULicense;

const
  nwkCmd_HEARTBEAT = 1;
  nwkCmd_RPT       = 2;
  nwkCmd_PTP       = 3;

  ccaiCmd_ZB_START_REQUEST        = $2600;
  ccaiCmd_ZB_BIND_DEVICE          = $2601;
  ccaiCmd_ZB_ALLOW_BIND           = $2602;
  ccaiCmd_ZB_SEND_DATA_REQUEST    = $2603;
  ccaiCmd_ZB_READ_CONFIGURATION   = $2604;
  ccaiCmd_ZB_WRITE_CONFIGURATION  = $2605;
  ccaiCmd_ZB_GET_DEVICE_INFO      = $2606;
  ccaiCmd_ZB_APP_REGISTER_REQUEST = $260A;
  ccaiCmd_SYS_RESET_IND           = $4180;
  ccaiCmd_ZB_START_CONFIRM        = $4680;
  ccaiCmd_ZB_BIND_CONFIRM         = $4681;
  ccaiCmd_ZB_ALLOW_BIND_CONFIRM   = $4682;
  ccaiCmd_ZB_SEND_DATA_CONFIRM    = $4683;
  ccaiCmd_ZB_RECEIVE_DATA_INDICATION = $4687;
  ccaiCmd_ZB_START_REQUEST_RSP    = $6600;
  ccaiCmd_ZB_ALLOW_BIND_RSP       = $6602;
  ccaiCmd_ZB_SEND_DATA_REQUEST_RSP = $6603;
  ccaiCmd_ZB_READ_CONFIGURATION_RSP = $6604;
  ccaiCmd_ZB_WRITE_CONFIGURATION_RSP = $6605;
  ccaiCmd_ZB_GET_DEVICE_INFO_RSP  = $6606;
  ccaiCmd_ZB_APP_REGISTER_REQUEST_RSP = $660A;

  szsCmd_INITIALIZATION       = $E100;
  szsCmd_LOCAL_SENSORS_REPORT = $E101;
  szsCmd_TEXT_OUTPUT          = $E102;
  szsCmd_MEMORY_DUMP          = $E103;
  szsCmd_CURRENTTIME          = $E104;

var
  stlSerialList:   TStringList;
  strRxBuf:        AnsiString;
  intPacketId:     Cardinal;
  TodayUnix:       Int64;
  currentFileName: String = '';
  prevDTsS:        QWord;
  prevDTsNs:       Cardinal;
  prevTS:          TDateTime;

{ TfrmMain }

procedure TfrmMain.actSetCoordExecute(Sender: TObject);
begin
  sendCommand(#225#16#0, Serial);
end;

procedure TfrmMain.actResetCC2480Execute(Sender: TObject);
begin
  sendCommand(#225#254, Serial);
end;

procedure TfrmMain.actMakeSinkExecute(Sender: TObject);
begin
  sendCommand(#225#17#1, Serial);
end;

procedure TfrmMain.actMemoryDumpExecute(Sender: TObject);
var
  strTmp: AnsiString;
begin
  strTmp := #225#253;
  strTmp += chr(StrToInt('$' + copy(mkeDumpAddress.Text, 3, 2)));
  strTmp += chr(StrToInt('$' + copy(mkeDumpAddress.Text, 1, 2)));
  strTmp += chr(speDumpSize.Value);
  sendCommand(strTmp, Serial);
end;

procedure TfrmMain.actConnOpenExecute(Sender: TObject);
begin
  if Serial.Active then
    actConnCloseExecute(Sender);
  if not Serial.Active then
  begin
    if (cmbSerialSelect.ItemIndex <> -1) then
      Serial.Device := stlSerialList.ValueFromIndex[cmbSerialSelect.ItemIndex];
    actEnumerateSerialExecute(Sender);
    if (cmbSerialSelect.ItemIndex <> -1) and (Serial.Device =
      stlSerialList.ValueFromIndex[cmbSerialSelect.ItemIndex]) then
    begin
      case cmbBaudRate.Text of
        '300': Serial.BaudRate    := br___300;
        '600': Serial.BaudRate    := br___600;
        '1200': Serial.BaudRate   := br__1200;
        '2400': Serial.BaudRate   := br__2400;
        '4800': Serial.BaudRate   := br__4800;
        '9600': Serial.BaudRate   := br__9600;
        '19200': Serial.BaudRate  := br_19200;
        '38400': Serial.BaudRate  := br_38400;
        '57600': Serial.BaudRate  := br_57600;
        '115200': Serial.BaudRate := br115200;
        '230400': Serial.BaudRate := br230400;
        '460800': Serial.BaudRate := br460800;
        '921600': Serial.BaudRate := br921600;
        else
          Serial.BaudRate := br__9600;
      end;
      case cmbDataBits.Text of
        '5': Serial.DataBits := db5bits;
        '6': Serial.DataBits := db6bits;
        '7': Serial.DataBits := db7bits;
        '8': Serial.DataBits := db8bits;
        else
          Serial.DataBits := db8bits;
      end;
      case cmbParity.Text of
        'None': serial.Parity  := pNone;
        'Space': Serial.Parity := pSpace;
        'Even': Serial.Parity  := pEven;
        'Odd': Serial.Parity   := pOdd;
        'Mark': Serial.Parity  := pMark;
        else
          serial.Parity := pNone;
      end;
      case cmbStopBits.Text of
        '1': serial.StopBits := sbOne;
        '2': Serial.StopBits := sbTwo;
        else
          serial.StopBits := sbOne;
      end;
      Serial.FlowControl := fcNone;
      try
        Serial.Open;
      except
        MessageDlg('Open serial port', 'The selected serial port could not be open.',
          mtError, [mbOK], 0);
      end;
      if Serial.Active then
      begin
        Application.Title := 'SZSM - ' +
          stlSerialList.ValueFromIndex[cmbSerialSelect.ItemIndex];
        Caption           := 'Simplified ZigBee Sensor Monitor - ' +
          stlSerialList.ValueFromIndex[cmbSerialSelect.ItemIndex];
        StatusBar1.Panels[0].Text :=
          'Connected: ' + stlSerialList.ValueFromIndex[cmbSerialSelect.ItemIndex] +
          ' - ' + cmbBaudRate.Text + ' ' + cmbDataBits.Text +
          Copy(cmbParity.Text, 1, 1) + cmbStopBits.Text;

        actConnOpen.Enabled     := False;
        actConnClose.Enabled    := True;
        actFileClean.Enabled    := True;
        actSetCoord.Enabled     := True;
        actSetRouter.Enabled    := True;
        actSetEndDevice.Enabled := True;
        actResetCC2480.Enabled  := True;
        actResetMSP430.Enabled  := True;
        actMakeSink.Enabled     := True;
        actSetClock.Enabled     := True;
        actMemoryDump.Enabled   := True;
      end;
    end;
  end;
end;

{$IFDEF MSWINDOWS}
procedure TfrmMain.actEnumerateSerialExecute(Sender: TObject);
{ modified code from Pato Díaz
 http://patotech.blogspot.com.br/2012/04/enumerate-com-ports-in-windows-with.html
 Quote from the page:
 "...
   Researching a bit to not reinvent the wheel, I found this page:
   http://www.lazarus.freepascal.org/index.php?topic=14313.0 which publishes
   three interesting functions.
   ...
   To solve my problem I took as example the first and the third function and
  made my own.
 "
 retrieved 2013-10-25 22:00:00
}
var
  reg: TRegistry;
  n:   Integer;

  function findFriendlyName(key: String; port: String): String;
  var
    r:  TRegistry;
    k:  TStringList;
    i:  Integer;
    ck: String;
    rs: String;
  begin
    r := TRegistry.Create;
    k := TStringList.Create;

    r.RootKey := HKEY_LOCAL_MACHINE;
    r.OpenKeyReadOnly(key);
    r.GetKeyNames(k);
    r.CloseKey;

    try
      for i := 0 to k.Count - 1 do
      begin
        ck := key + k[i] + '\'; // current key
        // looking for "PortName" stringvalue in "Device Parameters" subkey
        if r.OpenKeyReadOnly(ck + 'Device Parameters') then
        begin
          if r.ReadString('PortName') = port then
          begin
            r.CloseKey;
            r.OpenKeyReadOnly(ck);
            rs := r.ReadString('FriendlyName');
            if rs = '' then
              rs := port;
            Break;
          end; // if r.ReadString('PortName') = port ...
        end  // if r.OpenKeyReadOnly(ck + 'Device Parameters') ...
        // keep looking on subkeys for "PortName"
        else // if not r.OpenKeyReadOnly(ck + 'Device Parameters') ...
        if r.OpenKeyReadOnly(ck) and r.HasSubKeys then
        begin
          rs := findFriendlyName(ck, port);
          if rs <> '' then
            Break;
        end// if not (r.OpenKeyReadOnly(ck) and r.HasSubKeys) ...
        ; // if not r.OpenKeyReadOnly(ck + 'Device Parameters') ...
      end; // for i := 0 to k.Count - 1 ...
      Result := rs;
    finally
      r.Free;
      k.Free;
    end; // try ...
  end; // function findFriendlyName ...
begin
  reg := TRegistry.Create;

  try
    reg.RootKey := HKEY_LOCAL_MACHINE;
    if reg.OpenKeyReadOnly('HARDWARE\DEVICEMAP\SERIALCOMM') then
    begin
      cmbSerialSelect.Clear;
      reg.GetValueNames(stlSerialList);
      for n := 0 to stlSerialList.Count - 1 do
      begin
        stlSerialList[n] :=
          stlSerialList[n] + '=' + reg.ReadString(stlSerialList[n]);
        cmbSerialSelect.AddItem(stlSerialList.ValueFromIndex[n] +
          ' - ' + findFriendlyName('\System\CurrentControlSet\Enum\',
          stlSerialList.ValueFromIndex[n]), nil);
        if (n = 0) or (stlSerialList.ValueFromIndex[n] = Serial.Device) then
          cmbSerialSelect.ItemIndex := n;
      end; // for n := 0 to l.Count - 1 ...
    end // if reg.OpenKeyReadOnly('HARDWARE\DEVICEMAP\SERIALCOMM') ...
    else
      StatusBar1.Panels[0].Text := 'No serial port detected.';
    cmbSerialSelectChange(Sender);
  finally
    reg.Free;
  end; // try ...
end;

{$ELSE}
{ Modified from
 http://sourceforge.net/p/synalist/bugs/21/
 retrieved 2013-12-02 23:00:00

 Idea from PascalSCADA project's serialport.pas file:
 http://pascalscada.svn.sourceforge.net/viewvc/pascalscada/trunk/serialport.pas?revision=702&view=markup
 More prefixes exist see e.g. :
 http://comments.gmane.org/gmane.comp.ide.lazarus.general/46750
}
{$IFDEF UNIX}
procedure TfrmMain.actEnumerateSerialExecute(Sender: TObject);
var
{$IFDEF LINUX}
  // Possibly several other names too
  PortPrefix: array[0..2] of String = ('ttyS', 'ttyUSB', 'ttyACM');
{$ENDIF}{LINUX}
{$IFDEF FREEBSD}
  // http://www.freebsd.org/doc/handbook/serial.html
  PortPrefix: array[0..2] of String = ('cuad', 'cuau', 'ttyu');
{$ENDIF}{FREEBSD}
{$IFDEF NETBSD}
  // Don't have better info at the moment
  PortPrefix: array[0..0] of String = ('cuad');
{$ENDIF}{NETBSD}
{$IFDEF OPENBSD}
  // Don't have better info at the moment
  PortPrefix: array[0..0] of String = ('cuad');
{$ENDIF}{OPENBSD}
{$IFDEF Darwin}
  // Don't have better info at the moment
  PortPrefix: array[0..1] of String = ('cu.', 'tty.');
{$ENDIF}{Darwin}
  i:          Integer;
  sr:         TSearchRec;
begin
  cmbSerialSelect.ItemIndex := -1;
  cmbSerialSelect.Clear;
  stlSerialList.Clear;
  try
    for i := 0 to high(PortPrefix) do
      if FindFirst('/dev/' + PortPrefix[i] + '*', Longint($FFFFFFFF), sr) = 0 then
        // from patch
        repeat
          if (sr.Attr and $FFFFFFFF) = Sr.Attr then
          begin
            cmbSerialSelect.AddItem(sr.Name, nil);
            stlSerialList.Add(sr.Name + '=/dev/' + sr.Name);
            if Serial.Device = '/dev/' + sr.Name then
              cmbSerialSelect.ItemIndex := cmbSerialSelect.Items.Count - 1;
          end;
        until FindNext(sr) <> 0;
    cmbSerialSelectChange(Sender);
  finally
    FindClose(sr);
  end;
end;

{$ENDIF}{ UNIX }
{$ENDIF}{ MSWINDOWS }

procedure TfrmMain.actFileCleanExecute(Sender: TObject);
var
  strSensorList, strPTPList: AnsiString;
  intSensorIdx, intPTPIdx:   Integer;
begin
  intSensorIdx  := cmbSensorsDevice.ItemIndex;
  intPTPIdx     := cmbPTPTimingDevice.ItemIndex;
  strSensorList := cmbSensorsDevice.Items.Text;
  strPTPList    := cmbPTPTimingDevice.Items.Text;
  actFileNew.Execute;
  actConnOpen.Execute;
  cmbPTPTimingDevice.Items.Text := strPTPList;
  cmbSensorsDevice.Items.Text   := strSensorList;
  cmbPTPTimingDevice.ItemIndex  := intPTPIdx;
  cmbSensorsDevice.ItemIndex    := intSensorIdx;
end;

procedure TfrmMain.actFileExportAccept(Sender: TObject);
var
  DataSet: TDataSet;
  txtFile: TextFile;
  i:       Integer;
  s:       AnsiString;
begin
  DataSet := bdsPackets;
  DataSet.DisableControls;
  try
    { create a new file }
    AssignFile(txtFile, actFileExport.Dialog.FileName);
    try
      Rewrite(txtFile);
      s := ''; //initialize empty string
      { write field names (as column headers) }
      s := 'Id,TimeStamp,Length,Cmd,Command Name,Data';
      DataSet.First;
      { write field values }
      while not DataSet.EOF do
      begin
        { Numerical fields without quotes, string fields with quotes }
        s := s + #10;
        s := s + Format('%s,', [DataSet.FieldByName('Id').AsString]);
        s := s + Format('"%s",', [DataSet.FieldByName('TimeStamp').AsString]);
        s := s + Format('%s,', [DataSet.FieldByName('Length').AsString]);
        s := s + Format('"%s",', [DataSet.FieldByName('cmdHex').AsString]);
        s := s + Format('"%s",', [DataSet.FieldByName('cmdName').AsString]);
        s := s + Format('"%s"', [DataSet.FieldByName('DataHex').AsString]);
        if Length(s) >= 16384 then
        begin
          Write(txtFile, s);
          s := '';
        end;
        DataSet.Next;
      end;
      if Length(s) <> 0 then
      begin
        Write(txtFile, s);
        s := '';
      end;
    finally
      CloseFile(txtFile);
    end;
  finally
    DataSet.EnableControls;
  end;
  DataSet := nil;
end;

procedure TfrmMain.actFileNewExecute(Sender: TObject);
begin
  if Serial.Active then
    actConnCloseExecute(Sender);

  TodayUnix := DateTimeToUnix(Date);

  cmbSensorsDevice.Items.Clear;
  cmbSensorsDevice.AddItem('', nil);
  cmbSensorsDevice.ItemIndex := 0;

  currentFileName := '';
  bdsSensors.DisableControls;
  try
    bdsSensors.Filtered := False;
    bdsSensors.First;
    while not bdsSensors.EOF do
      bdsSensors.Delete;
  finally
    bdsSensors.EnableControls;
  end;

  cmbPTPTimingDevice.Items.Clear;
  cmbPTPTimingDevice.AddItem('', nil);
  cmbPTPTimingDevice.ItemIndex := 0;

  bdsPTPTiming.DisableControls;
  try
    bdsPTPTiming.Filtered := False;
    bdsPTPTiming.First;
    while not bdsPTPTiming.EOF do
      bdsPTPTiming.Delete;
  finally
    bdsPTPTiming.EnableControls;
  end;

  bdsPackets.DisableControls;
  try
    bdsPackets.Filtered := False;
    bdsPackets.First;
    while not bdsPackets.EOF do
      bdsPackets.Delete;
  finally
    bdsPackets.EnableControls;
  end;
  intPacketId := 1;

  mmoTextRcv.Clear;

  //{ Test Data: }

  //// Allow bind Packet
  //bdsPackets.Append;
  //bdsPackets['Length']  := 1;
  //bdsPackets['Command'] := ccaiCmd_ZB_ALLOW_BIND;
  //bdsPackets.FieldByName('Data').AsString := #255;
  //bdsPackets.Post;

  //// Local sensor report
  //bdsPackets.Append;
  //bdsPackets['Length']  := 76;
  //bdsPackets['Command'] := szsCmd_LOCAL_SENSORS_REPORT;
  //bdsPackets.FieldByName('Data').AsString :=
  //  #9#7#6#5#4#3#2#1 + // device address
  //  #141#2#53#0#0#0#0#0#64#130#81#15 + // Temperature
  //  #195#3#53#0#0#0#0#0#0#100#54#16 + // Voltage
  //  #115#0#53#0#0#0#0#0#192#69#27#17 + // Luminance
  //  #255#255#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0; // Clock adjust
  //bdsPackets.Post;

  //// Local sensor report
  //bdsPackets.Append;
  //bdsPackets['Length']  := 76;
  //bdsPackets['Command'] := szsCmd_LOCAL_SENSORS_REPORT;
  //bdsPackets.FieldByName('Data').AsString :=
  //  #9#7#6#5#4#3#2#1 + // device address
  //  #241#2#57#0#0#0#0#0#64#130#81#15 + // Temperature
  //  #235#3#57#0#0#0#0#0#0#100#54#16 + // Voltage
  //  #135#250#57#0#0#0#0#0#192#69#27#17 + // Luminance
  //  #255#255#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0; // Clock adjust
  //bdsPackets.Post;

  //// Remote sensor report
  //bdsPackets.Append;
  //bdsPackets['Length']  := 82;
  //bdsPackets['Command'] := ccaiCmd_ZB_RECEIVE_DATA_INDICATION;
  //bdsPackets.FieldByName('Data').AsString :=
  //  #1#0#2#0#76#0 + //src, cmd, size
  //  #8#7#6#5#4#3#2#1 + // device address
  //  #181#2#52#0#0#0#0#0#64#130#81#15 + // Temperature
  //  #205#3#52#0#0#0#0#0#0#100#54#16 + // Voltage
  //  #105#0#52#0#0#0#0#0#192#69#27#17 + // Luminance
  //  #255#255#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0; // Clock adjust
  //bdsPackets.Post;

  //// Remote sensor report
  //bdsPackets.Append;
  //bdsPackets['Length']  := 82;
  //bdsPackets['Command'] := ccaiCmd_ZB_RECEIVE_DATA_INDICATION;
  //bdsPackets.FieldByName('Data').AsString :=
  //  #1#0#2#0#76#0 + //src, cmd, size
  //  #8#7#6#5#4#3#2#1 + // device address
  //  #221#2#56#0#0#0#0#0#64#130#81#15 + // Temperature
  //  #255#3#56#0#0#0#0#0#0#100#54#16 + // Voltage
  //  #165#0#56#0#0#0#0#0#192#69#27#17 + // Luminance
  //  #255#255#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0#0; // Clock adjust
  //bdsPackets.Post;
  //{ End of Test Data }

end;

procedure TfrmMain.actFileOpenAccept(Sender: TObject);
var
  fs: TFileStream;
  rs: TResourceStream;
  vr: TVersionResource;
  fi: TVersionFixedInfo;
begin
  currentFileName := actFileOpen.Dialog.FileName;
  fs := TFileStream.Create(currentFileName, fmOpenRead);
  try
    fs.Seek(0, soFromBeginning);
    { Read Header begin }
    if fs.ReadAnsiString <> 'SZSM' then
      raise EFormatError.Create('Invalid file header');

    { Read program version }
    { based on:
      http://wiki.lazarus.freepascal.org/Show_Application_Title,_Version,_and_Company }
    rs := TResourceStream.CreateFromID(HINSTANCE, 1, PChar(RT_VERSION));
    try
      vr := TVersionResource.Create;
      try
        vr.SetCustomRawDataStream(rs);
        fi := vr.FixedInfo;
        try
          if fs.ReadWord <> fi.ProductVersion[0] then
            raise EFormatError.Create('File version doesn''t match program version');
          fs.ReadWord;
        finally
          fi := nil
        end;
        vr.SetCustomRawDataStream(nil);
      finally
        vr.Free;
      end;
    finally
      rs.Free;
    end;

    { Clear current data }
    actFileNew.Execute;

    { load current autoInc Value }
    intPacketId := fs.ReadDWord;

    { load data }
    bdsPackets.DisableControls;
    try
      bdsPackets.LoadFromStreamAt(fs);
      bdsPackets.Active := True;
    finally
      bdsPackets.EnableControls;
    end;
  finally
    fs.Free;
  end;
end;

procedure TfrmMain.actFileSaveAsAccept(Sender: TObject);
begin
  currentFileName := actFileSaveAs.Dialog.FileName;
  ActionList1.ActionByName('actFileSave').Execute;
end;

procedure TfrmMain.actFileSaveAsBeforeExecute(Sender: TObject);
begin
  if currentFileName <> '' then
    actFileSaveAs.Dialog.FileName := currentFileName;
end;

procedure TfrmMain.actFileSaveExecute(Sender: TObject);
var
  fs: TFileStream;
  rs: TResourceStream;
  vr: TVersionResource;
  fi: TVersionFixedInfo;
begin
  if currentFileName = '' then
    ActionList1.ActionByName('actFileSaveAs').Execute
  else if FileExists(currentFileName) and not FileIsWritable(currentFileName) then

    ActionList1.ActionByName('actFileSaveAs').Execute
  else
  begin
    fs := TFileStream.Create(currentFileName, fmCreate);
    try
      fs.Seek(0, soFromBeginning);
      { Save Header begin }
      fs.WriteAnsiString('SZSM');
      { Save program version }
      { based on:
        http://wiki.lazarus.freepascal.org/Show_Application_Title,_Version,_and_Company }
      rs := TResourceStream.CreateFromID(HINSTANCE, 1, PChar(RT_VERSION));
      try
        vr := TVersionResource.Create;
        try
          vr.SetCustomRawDataStream(rs);
          fi := vr.FixedInfo;
          try
            fs.WriteWord(fi.ProductVersion[0]);
            fs.WriteWord(fi.ProductVersion[1]);
          finally
            fi := nil
          end;
          vr.SetCustomRawDataStream(nil);
        finally
          vr.Free;
        end;
      finally
        rs.Free;
      end;
      { save current autoInc Value }
      fs.WriteDWord(intPacketId);
      { save data }
      bdsPackets.AppendToStream(fs);
    finally
      fs.Free;
    end;
  end;
end;

procedure TfrmMain.actConnCloseExecute(Sender: TObject);
begin
  actConnClose.Enabled    := False;
  actFileClean.Enabled    := False;
  actSetCoord.Enabled     := False;
  actSetRouter.Enabled    := False;
  actSetEndDevice.Enabled := False;
  actResetCC2480.Enabled  := False;
  actResetMSP430.Enabled  := False;
  actMakeSink.Enabled     := False;
  actSetClock.Enabled     := False;
  actMemoryDump.Enabled   := False;
  Serial.Close;
  cmbSerialSelectChange(Self);

  StatusBar1.Panels[0].Text := 'No connection';
  Caption           := 'Simplified ZigBee Sensor Monitor';
  Application.Title := 'SZSM';
end;

procedure TfrmMain.actResetMSP430Execute(Sender: TObject);
begin
  sendCommand(#225#255, Serial);
end;

procedure TfrmMain.actSetClockExecute(Sender: TObject);
var
  pUInt64Tmp:  PQWord;
  pUInt32Tmp:  PCardinal;
  currentDate: TDateTime;
  UnixTS:      QWord;
  UnixTSNs:    Cardinal;
  cmd:         ShortString;
begin
  currentDate := Now;
  UnixTS      := DateTimeToUnix(currentDate);
  UnixTSNs    := round(frac(currentDate * 24 * 3600) * 1000000000);
  cmd         := #225#32;
  SetLength(cmd, 12);
  pUInt64Tmp  := @cmd[3];
  pUInt64Tmp^ := UnixTS;
  pUInt64Tmp  := nil;
  pUInt32Tmp  := @cmd[9];
  pUInt32Tmp^ := UnixTSNs;
  pUInt32Tmp  := nil;
  sendCommand(cmd, Serial);
  prevDTsS  := 0;
  prevDTsNs := 0;
  prevTS    := Now;
end;

procedure TfrmMain.actSetEndDeviceExecute(Sender: TObject);
begin
  sendCommand(#225#16#2, Serial);
end;

procedure TfrmMain.actSetRouterExecute(Sender: TObject);
begin
  sendCommand(#225#16#1, Serial);
end;

procedure TfrmMain.actViewConnPanelExecute(Sender: TObject);
begin
  pnlConnSetup.Visible := actViewConnPanel.Checked;
end;

procedure TfrmMain.bdsPacketsAfterInsert(DataSet: TDataSet);
begin
  with DataSet do
    if FieldByName('Id').IsNull then
    begin
      FieldByName('Id').ReadOnly  := False;
      FieldByName('Id').AsInteger := intPacketId;
      FieldByName('Id').ReadOnly  := True;
      Inc(intPacketId);
    end;
end;

procedure TfrmMain.bdsPacketsAfterPost(DataSet: TDataSet);
var
  pUInt16Tmp: PWord;
  pUInt64SrcIeeeAddr: PQWord;
  strTmp: String;
  bm:     TBookmarkStr; // Deprecated infavor of TBookmark, but DataSet still uses it
begin
  with DataSet do
  begin
    if FieldByName('Id').IsNull or FieldByName('Length').IsNull or
      FieldByName('Command').IsNull or FieldByName('Data').IsNull then
      exit;

    { test for sensor packet from network }
    if (FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) or
      (FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION) or
      (FieldByName('Command').AsInteger = szsCmd_LOCAL_SENSORS_REPORT) then
    begin
      strTmp := FieldByName('Data').AsString;
      if (FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) then
      begin
        pUInt16Tmp := @strTmp[1 + 2]; // packet command
        if pUInt16Tmp^ <> nwkCmd_RPT then
          Exit;
        pUInt64SrcIeeeAddr := @strTmp[1 + 8]; // Source IEEE address
      end
      else if (FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION) then
      begin
        pUInt16Tmp := @strTmp[1 + 2]; // packet command
        if pUInt16Tmp^ <> nwkCmd_RPT then
          Exit;
        pUInt64SrcIeeeAddr := @strTmp[1 + 6]; // Source IEEE address
      end
      else
        pUInt64SrcIeeeAddr := @strTmp[1]; // Source IEEE address

      { Verify if address of packet source is on the dropdown list }
      if cmbSensorsDevice.Items.IndexOf(IntToHex(pUInt64SrcIeeeAddr^, 16)) = -1 then
        cmbSensorsDevice.AddItem(IntToHex(pUInt64SrcIeeeAddr^, 16), nil);
      if cmbSensorsDevice.Text = IntToHex(pUInt64SrcIeeeAddr^, 16) then
      begin
        bdsSensors.DisableControls;
        if not cbRollSensorList.Checked then
          bm := bdsSensors.Bookmark;
        try
          copyPacketToGraphSensors(DataSet, bdsSensors, cmbSensorsDevice.Text);
        finally
          if not cbRollSensorList.Checked then
            try
              bdsSensors.Bookmark := bm;
            finally
            end;
          bdsSensors.EnableControls;
        end;
      end;

      { Only add on PTP graph if adjust time is <> 0 }

      { Verify if address of packet source is on the dropdown list }
      if cmbPTPTimingDevice.Items.IndexOf(IntToHex(pUInt64SrcIeeeAddr^, 16)) = -1 then
        cmbPTPTimingDevice.AddItem(IntToHex(pUInt64SrcIeeeAddr^, 16), nil);
      if cmbPTPTimingDevice.Text = IntToHex(pUInt64SrcIeeeAddr^, 16) then
      begin
        bdsPTPTiming.DisableControls;
        if not cbRollPtpList.Checked then
          bm := bdsPTPTiming.Bookmark;
        try
          copyPacketToGraphPTP(DataSet, bdsPTPTiming, cmbPTPTimingDevice.Text);
        finally
          if not cbRollPtpList.Checked then
            try
              bdsPTPTiming.Bookmark := bm;
            finally
            end;
          bdsPTPTiming.EnableControls;
        end;
      end;
    end;
  end;
end;

procedure TfrmMain.bdsPacketsCalcFields(DataSet: TDataSet);
var
  i, n:       Byte;
  TS:         TDateTime;
  CmdBytes:   TBytes;
  CmdHexText: String;
begin
  with DataSet do
  begin
    { TimeStamp }
    if (not FieldByName('TimeStampUnix').IsNull) then
    begin
      TS := UnixToDateTime(FieldByName('TimeStampUnix').AsLargeInt);
      if (not FieldByName('TimeStampMs').IsNull) then
        TS := TS + (FieldByName('TimeStampMs').AsInteger / MSecsPerDay);
      FieldByName('TimeStamp').AsDateTime := TS;
    end;

    { Hex representation of Command }
    if (not FieldByName('Command').IsNull) then
    begin
      CmdHexText := IntToHex(FieldByName('Command').AsInteger, 4);
      FieldByName('cmdHex').AsString := CmdHexText;
    end;

    { Hex representation of Data }
    if (not FieldByName('Length').IsNull) and (not FieldByName('Data').IsNull) then
    begin
      n := FieldByName('Length').AsInteger;
      if n <> 0 then
      begin
        CmdBytes   := FieldByName('Data').AsBytes;
        CmdHexText := IntToHex(CmdBytes[0], 2);
        Dec(n);
        for i := 1 to n do
          AppendStr(CmdHexText, ' ' + IntToHex(CmdBytes[i], 2));
        FieldByName('DataHex').AsString := CmdHexText;
      end;
    end;
  end;
end;

procedure TfrmMain.bdsPTPTimingCalcFields(DataSet: TDataSet);
begin
  with DataSet do
    { time of PTP adjust }
    if not (FieldByName('PTPTimeS').IsNull or FieldByName('PTPTimeNS').IsNull) then
      if FieldByName('PTPTimeS').AsInteger > TodayUnix then
        FieldByName('PTPTime').AsFloat :=
          (FieldByName('PTPTimeS').AsInteger +
          (FieldByName('PTPTimeNS').AsInteger / 1000000000.0)) - TodayUnix
      else
        FieldByName('PTPTime').AsFloat :=
          FieldByName('PTPTimeS').AsInteger +
          (FieldByName('PTPTimeNS').AsInteger / 1000000000.0);

end;

procedure TfrmMain.bdsSensorsCalcFields(DataSet: TDataSet);
begin
  with DataSet do
  begin
    { Temperature time }
    if not (FieldByName('TempTimeS').IsNull or FieldByName('TempTimeNS').IsNull) then
      if FieldByName('TempTimeS').AsInteger > TodayUnix then
        FieldByName('TempTime').AsFloat :=
          (FieldByName('TempTimeS').AsInteger +
          (FieldByName('TempTimeNS').AsInteger / 1000000000.0)) - TodayUnix
      else
        FieldByName('TempTime').AsFloat :=
          FieldByName('TempTimeS').AsInteger +
          (FieldByName('TempTimeNS').AsInteger / 1000000000.0);

    { Voltage time }
    if not (FieldByName('VoltTimeS').IsNull or FieldByName('VoltTimeNS').IsNull) then
      if FieldByName('VoltTimeS').AsInteger > TodayUnix then
        FieldByName('VoltTime').AsFloat :=
          (FieldByName('VoltTimeS').AsInteger +
          (FieldByName('VoltTimeNS').AsInteger / 1000000000.0)) - TodayUnix
      else
        FieldByName('VoltTime').AsFloat :=
          FieldByName('VoltTimeS').AsInteger +
          (FieldByName('VoltTimeNS').AsInteger / 1000000000.0);

    { Luminance time }
    if not (FieldByName('LumTimeS').IsNull or FieldByName('LumTimeNS').IsNull) then
      if FieldByName('LumTimeS').AsInteger > TodayUnix then
        FieldByName('LumTime').AsFloat :=
          (FieldByName('LumTimeS').AsInteger +
          (FieldByName('LumTimeNS').AsInteger / 1000000000.0)) - TodayUnix
      else
        FieldByName('LumTime').AsFloat :=
          FieldByName('LumTimeS').AsInteger +
          (FieldByName('LumTimeNS').AsInteger / 1000000000.0);
  end;
end;

procedure TfrmMain.btnPTPExportClick(Sender: TObject);
var
  DataSet: TDataSet;
  txtFile: TextFile;
  i:       Integer;
  s:       AnsiString;
begin
  SaveDialog.Title   := 'Save CSV file from PTP timing data as';
  SaveDialog.Options := [ofOverwritePrompt, ofNoReadOnlyReturn,
    ofEnableSizing, ofViewDetail];
  SaveDialog.Filter  := 'Comma separated value (*.csv)|*.csv|All files (*.*)|*.*';
  if SaveDialog.Execute then
  begin
    DataSet := bdsPTPTiming;
    DataSet.DisableControls;
    try
      { create a new file }
      AssignFile(txtFile, SaveDialog.FileName);
      try
        Rewrite(txtFile);
        s := ''; //initialize empty string
        { write field names (as column headers) }
        s := 'Id,Device,PTP Time (s),Delay to Master (ns),Last Adjust (ns)';
        DataSet.First;
        { write field values }
        while not DataSet.EOF do
        begin
          { Numerical fields without quotes, string fields with quotes }
          s := s + #10;
          s := s + Format('%s,', [DataSet.FieldByName('Id').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('DeviceHex').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('PTPTime').AsString]);
          s := s + Format('%s,', [DataSet.FieldByName('DelayToMaster').AsString]);
          s := s + Format('%s', [DataSet.FieldByName('LastAdjust').AsString]);
          if Length(s) >= 16384 then
          begin
            Write(txtFile, s);
            s := '';
          end;
          DataSet.Next;
        end;
        if Length(s) <> 0 then
        begin
          Write(txtFile, s);
          s := '';
        end;
      finally
        CloseFile(txtFile);
      end;
    finally
      DataSet.EnableControls;
    end;
    DataSet := nil;
  end;
end;

procedure TfrmMain.btnSensorExportClick(Sender: TObject);
var
  DataSet: TDataSet;
  txtFile: TextFile;
  i:       Integer;
  s:       AnsiString;
begin
  SaveDialog.Title   := 'Save CSV file from Sensors data as';
  SaveDialog.Options := [ofOverwritePrompt, ofNoReadOnlyReturn,
    ofEnableSizing, ofViewDetail];
  SaveDialog.Filter  := 'Comma separated value (*.csv)|*.csv|All files (*.*)|*.*';
  if SaveDialog.Execute then
  begin
    DataSet := bdsSensors;
    DataSet.DisableControls;
    try
      { create a new file }
      AssignFile(txtFile, SaveDialog.FileName);
      try
        Rewrite(txtFile);
        s := ''; //initialize empty string
        { write field names (as column headers) }
        s :=
          'Id,Device,Temp. Time (s),Temp. Value (°C),Volt. Time (s),Volt. Value (V),Lum. Time (s),Lum. Value (unk)';
        DataSet.First;
        { write field values }
        while not DataSet.EOF do
        begin
          { Numerical fields without quotes, string fields with quotes }
          s := s + #10;
          s := s + Format('%s,', [DataSet.FieldByName('Id').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('DeviceHex').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('TempTime').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('TempValue').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('VoltTime').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('VoltValue').AsString]);
          s := s + Format('"%s",', [DataSet.FieldByName('LumTime').AsString]);
          s := s + Format('"%s"', [DataSet.FieldByName('LumValue').AsString]);
          if Length(s) >= 16384 then
          begin
            Write(txtFile, s);
            s := '';
          end;
          DataSet.Next;
        end;
        if Length(s) <> 0 then
        begin
          Write(txtFile, s);
          s := '';
        end;
      finally
        CloseFile(txtFile);
      end;
    finally
      DataSet.EnableControls;
    end;
    DataSet := nil;
  end;
end;

procedure TfrmMain.cmbPTPTimingDeviceChange(Sender: TObject);
begin
  copyPacketsToGraphPTP(bdsPackets, bdsPTPTiming, cmbPTPTimingDevice);
end;

procedure TfrmMain.cmbSensorsDeviceChange(Sender: TObject);
begin
  copyPacketsToGraphSensors(bdsPackets, bdsSensors, cmbSensorsDevice);
end;

procedure TfrmMain.cmbSerialSelectChange(Sender: TObject);
begin
  actConnOpen.Enabled := not Serial.Active and (cmbSerialSelect.ItemIndex <> -1);
end;

procedure TfrmMain.dbgPacketsCellClick(Column: TColumn);
begin
  showPacket(dbgPackets.DataSource.DataSet, mmoTextRcv);
end;

procedure TfrmMain.dbgPacketsKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key in [VK_UP, VK_DOWN] then
    showPacket(dbgPackets.DataSource.DataSet, mmoTextRcv)
  else if key = VK_HOME then
  begin
    dbgPackets.DataSource.DataSet.First;
    showPacket(dbgPackets.DataSource.DataSet, mmoTextRcv);
  end
  else if key = VK_END then
  begin
    dbgPackets.DataSource.DataSet.Last;
    showPacket(dbgPackets.DataSource.DataSet, mmoTextRcv);
  end;
end;

procedure TfrmMain.dbgPTPTimingKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key = VK_HOME then
    dbgPackets.DataSource.DataSet.First
  else if key = VK_END then
    dbgPackets.DataSource.DataSet.Last;
end;

procedure TfrmMain.dbgSensorsKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key = VK_HOME then
    dbgPackets.DataSource.DataSet.First
  else if key = VK_END then
    dbgPackets.DataSource.DataSet.Last;
end;

procedure TfrmMain.dcsSensorsTempGetItem(ASender: TDbChartSource;
  var AItem: TChartDataItem);
begin
end;

procedure TfrmMain.FormActivate(Sender: TObject);
begin
  if stlSerialList.Count = 0 then
    actEnumerateSerialExecute(Sender);
end;

procedure TfrmMain.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  Serial.Active := False;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  {$IFDEF Darwin}
  Color := clBtnFace;
  btnOpenSerial.Height := 22;
  btnCloseSerial.Height := 22;
  {$ENDIF}

  stlSerialList := TStringList.Create;
  strRxBuf      := '';
  intPacketId   := 1;

  prevDTsS  := 0;
  prevDTsNs := 0;
  prevTS    := Now;

  TodayUnix := DateTimeToUnix(Date);

  { Ensure that datasets are opened }
  bdsCommands.Open;
  bdsPackets.Open;
  bdsSensors.Open;
  bdsPTPTiming.Open;

  { Fill commands dataset }
  FillmdsCommands(bdsCommands);

  actFileNew.Execute;

end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  stlSerialList.Free;
end;

procedure TfrmMain.actPacketsCopyLineExecute(Sender: TObject);
var
  strTmp: String;
begin
  if bdsPackets.FieldByName('Id').IsNull then
    Exit;

  strTmp := bdsPackets.FieldByName('Id').AsString + #9;
  if not bdsPackets.FieldByName('Length').IsNull then
    strTmp := strTmp + bdsPackets.FieldByName('Length').AsString;
  strTmp   := strTmp + #9;
  if not bdsPackets.FieldByName('cmdHex').IsNull then
    strTmp := strTmp + bdsPackets.FieldByName('cmdHex').AsString;
  strTmp   := strTmp + #9;
  if not bdsPackets.FieldByName('cmdName').IsNull then
    strTmp := strTmp + bdsPackets.FieldByName('cmdName').AsString;
  strTmp   := strTmp + #9;
  if not bdsPackets.FieldByName('DataHex').IsNull then
    strTmp         := strTmp + bdsPackets.FieldByName('DataHex').AsString;
  Clipboard.AsText := strTmp;
end;

procedure TfrmMain.actHelpLicenseExecute(Sender: TObject);
begin
  if frmLicense = nil then
    frmLicense := TfrmLicense.Create(Self);
  frmLicense.ShowModal;
  FreeAndNil(frmLicense);
end;

procedure TfrmMain.FormDropFiles(Sender: TObject; const FileNames: array of String);
var
  testFileName, FileName: String;
begin
  FileName := '';
  for testFileName in FileNames do
    if ExtractFileExt(testFileName) = '.szs' then
      FileName := testFileName;
  if FileName <> '' then
  begin
    actFileOpen.Dialog.FileName := FileName;
    actFileOpenAccept(Sender);
  end;
end;

procedure TfrmMain.mkeDumpAddressExit(Sender: TObject);
var
  i:      Byte;
  strTmp: AnsiString;
begin
  strTmp := mkeDumpAddress.EditText;
  for i := 1 to Length(strTmp) do
    if not (strTmp[i] in ['0'..'9', 'A'..'F']) then
      strTmp[i] := '0';
  if mkeDumpAddress.EditText <> strTmp then
    mkeDumpAddress.EditText := strTmp;
end;

procedure TfrmMain.mkeDumpAddressKeyPress(Sender: TObject; var Key: Char);
begin
  if key in ['a' .. 'f'] then
    key := chr(Ord(key) - 32)
  else if not (Key in ['0'..'9', 'A'..'F', #8]) then
    Key := #0;
end;

procedure TfrmMain.SerialRxData(Sender: TObject);
begin
  strRxBuf := strRxBuf + Serial.ReadData;
  processPacket(strRxBuf, bdsPackets, not cbRollPacketList.Checked);
end;

{ Others }

procedure copyPacketToGraphPTP(pDS, ptpDS: TDataSet; device: String);
var
  buf:        AnsiString;
  pBuf:       PByte;
  strPktSrc:  String;
  pktLen:     Word;
  vTsS:       QWord; // Time stamp of a value, seconds part
  vTsNs:      Cardinal; // Time stamp of a value, Nanoseconds part
  pbytTmp:    PByte;
  pUInt16Tmp: PWord;
  PUInt32Tmp: PCardinal;
  PUInt64Tmp: PQWord;
  PInt64Tmp:  PInt64;
  i:          Byte;
begin
  vTsS := 0;
  buf  := pDS.FieldByName('Data').AsString;
  pBuf := @buf[1];

  if pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST then
    // Simple API command
  begin
    pktLen := pDS.FieldByName('Length').AsInteger - 8;

    if pktLen <> pBuf[7] then
      Exit;

    pUInt16Tmp := @pBuf[2]; // Packet Command
    if pUInt16Tmp^ <> nwkCmd_RPT then
      Exit;

    pBuf := pBuf + 8; // Go to start of report
  end
  else if pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION then
    // Simple API command
  begin
    pktLen     := pDS.FieldByName('Length').AsInteger - 6;
    pUInt16Tmp := @pBuf[4];
    if pktLen <> pUInt16Tmp^ then
      Exit;

    pUInt16Tmp := @pBuf[2]; // Packet Command
    if pUInt16Tmp^ <> nwkCmd_RPT then
      Exit;

    pBuf := pBuf + 6; // Go to start of report
  end
  else if pDS.FieldByName('Command').AsInteger = szsCmd_LOCAL_SENSORS_REPORT then
    pktLen := pDS.FieldByName('Length').AsInteger
  else
    Exit;

  if (pktLen <> 76) then
    Exit;

  // Check if src is from the required device
  PUInt64Tmp := @pBuf[0];
  strPktSrc  := IntToHex(PUInt64Tmp^, 16);
  if strPktSrc <> device then
    Exit;

  pbytTmp := @vTsS; // point to dest variable,
  for i := 0 to 5 do // for each byte, copy value from source to dest (don't go beyond)
    pbytTmp[i] := pBuf[62 + i];
  PUInt32Tmp   := @pBuf[68];
  vTsNs        := PUInt32Tmp^;

  { ignore if adjust time is 0
    no adjust were done but values are included in the report }
  if vTsS = 0 then
    Exit;

  { check number of lines and don't let it to go over 1000 data points }
  while ptpDS.RecordCount >= 1000 do
  begin
    ptpDS.First;
    ptpDS.Delete;
  end;

  { Insert on graph table }
  ptpDS.Append;
  ptpDS.FieldByName('Id').AsInteger := pDS.FieldByName('Id').AsInteger;
  ptpDS.FieldByName('DeviceHex').AsString := strPktSrc;
  PUInt64Tmp := @pBuf[46];
  ptpDS.FieldByName('DelayToMaster').AsFloat := PUInt64Tmp^;
  PInt64Tmp  := @pBuf[54];
  ptpDS.FieldByName('LastAdjust').AsLargeInt := PInt64Tmp^;
  ptpDS.FieldByName('PTPTimeS').AsLargeInt := vTsS;
  ptpDS.FieldByName('PTPTimeNS').AsInteger := vTsNs;
  ptpDS.Post;
end;

procedure copyPacketToGraphSensors(pDS, sDS: TDataSet; device: String);
var
  i:          Integer;
  buf:        AnsiString;
  pBuf:       PByte;
  strPktSrc:  String;
  pktLen:     Word;
  v:          Word; // data value on a report
  vTsS:       QWord; // Time stamp of a value, seconds part
  vTsNs:      Cardinal; // Time stamp of a value, Nanoseconds part
  pUInt16Tmp: PWord;
  PUInt32Tmp: PCardinal;
  PUInt64Tmp: PQWord;
  pbytTmp:    PByte;
begin
  vTsS := 0;
  buf  := pDS.FieldByName('Data').AsString;
  pBuf := @buf[1];

  if pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST then
    // Simple API command send
  begin
    pktLen := pDS.FieldByName('Length').AsInteger - 8;
    if pktLen <> pBuf[7] then
      Exit;

    pUInt16Tmp := @pBuf[2]; // Packet Command
    if pUInt16Tmp^ <> nwkCmd_RPT then
      Exit;

    pBuf := pBuf + 8; // Go to start of report
  end
  else if pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION then
    // Simple API command receive
  begin
    pktLen     := pDS.FieldByName('Length').AsInteger - 6;
    pUInt16Tmp := @pBuf[4];
    if pktLen <> pUInt16Tmp^ then
      Exit;

    pUInt16Tmp := @pBuf[2]; // Packet Command
    if pUInt16Tmp^ <> nwkCmd_RPT then
      Exit;

    pBuf := pBuf + 6; // Go to start of report
  end
  else if pDS.FieldByName('Command').AsInteger = szsCmd_LOCAL_SENSORS_REPORT then
    pktLen := pDS.FieldByName('Length').AsInteger
  else
    Exit;

  if ((pktLen <> 44) and (pktLen <> 76)) then
    Exit;

  PUInt64Tmp := @pBuf[0];
  strPktSrc  := IntToHex(PUInt64Tmp^, 16);
  if strPktSrc <> device then
    Exit;

  { check number of lines and don't let to go over 1000 data points }
  while sDS.RecordCount >= 1000 do
  begin
    sDS.First;
    sDS.Delete;
  end;

  { Insert on graph table }
  sDS.Append;
  sDS.FieldByName('Id').AsInteger := pDS.FieldByName('Id').AsInteger;
  { Device unique address }
  sDS.FieldByName('DeviceHex').AsString := strPktSrc;
  { Temperature }
  pUInt16Tmp := @pBuf[8];
  v          := pUInt16Tmp^;
  pbytTmp    := @vTsS; // point to dest variable,
  for i := 0 to 5 do // for each byte, copy value from source to dest (don't go beyond)
    pbytTmp[i] := pBuf[10 + i];
  PUInt32Tmp := @pBuf[10 + 6];
  vTsNs   := PUInt32Tmp^;
  sDS.FieldByName('TempTimeS').AsLargeInt := vTsS;
  sDS.FieldByName('TempTimeNS').AsInteger := vTsNs;
  sDS.FieldByName('TempValue').AsFloat := 0.4126320423 * v - 277.746479;
  // refer to showSensorReport
  { Voltage }
  pUInt16Tmp := @pBuf[20];
  v       := pUInt16Tmp^;
  pbytTmp := @vTsS; // point to dest variable,
  for i := 0 to 5 do // for each byte, copy value from source
    pbytTmp[i] := pBuf[22 + i]; // to dest (don't go beyond)
  PUInt32Tmp := @pBuf[22 + 6];
  vTsNs   := PUInt32Tmp^;
  sDS.FieldByName('VoltTimeS').AsLargeInt := vTsS;
  sDS.FieldByName('VoltTimeNS').AsInteger := vTsNs;
  sDS.FieldByName('VoltValue').AsFloat := v / 341; // refer to showSensorReport
  { Luminance }
  pUInt16Tmp := @pBuf[32];
  v       := pUInt16Tmp^;
  pbytTmp := @vTsS; // point to dest variable,
  for i := 0 to 5 do // for each byte, copy value from source to dest (don't go beyond)
    pbytTmp[i] := pBuf[34 + i];
  PUInt32Tmp   := @pBuf[34 + 6];
  vTsNs        := PUInt32Tmp^;
  sDS.FieldByName('LumTimeS').AsLargeInt := vTsS;
  sDS.FieldByName('LumTimeNS').AsInteger := vTsNs;
  sDS.FieldByName('LumValue').AsFloat := v; // refer to showSensorReport
  sDS.Post;
end;

procedure copyPacketsToGraphPTP(pDS, ptpDS: TDataSet; cmb: TComboBox);
var
  bm:         TBookmarkStr; // Deprecated infavor of TBookmark, but DataSet still uses it
  strTmp:     String;
  pUInt16Tmp: PWord;
  pUInt64Tmp: PQWord;
  nwkCmd:     Word;
begin
  ptpDS.DisableControls;
  try
    ptpDS.First;
    while not ptpDS.EOF do
      ptpDS.Delete;

    pDS.DisableControls;
    bm := pDS.Bookmark;
    try
      pDS.First;
      while not pDS.EOF do
      begin
        if not (pDS.FieldByName('Id').IsNull or pDS.FieldByName('Length').IsNull or
          pDS.FieldByName('Command').IsNull or pDS.FieldByName('Data').IsNull) then
          { test for sensor packet from network }
          if (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) or
            (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION) or
            (pDS.FieldByName('Command').AsInteger = szsCmd_LOCAL_SENSORS_REPORT) then
          begin
            strTmp := pDS.FieldByName('Data').AsString;
            { Verify if address of packet source is on the dropdown list }
            if (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) then
            begin
              pUInt16Tmp := @strTmp[3];
              nwkCmd     := pUInt16Tmp^;
              pUInt64Tmp := @strTmp[1 + 8];
            end
            else if (pDS.FieldByName('Command').AsInteger =
              ccaiCmd_ZB_RECEIVE_DATA_INDICATION) then
            begin
              pUInt16Tmp := @strTmp[3];
              nwkCmd     := pUInt16Tmp^;
              pUInt64Tmp := @strTmp[1 + 6];
            end
            else
            begin
              nwkCmd     := nwkCmd_RPT;
              pUInt64Tmp := @strTmp[1];
            end;
            if nwkCmd = nwkCmd_RPT then
            begin
              if cmb.Items.IndexOf(IntToHex(pUInt64Tmp^, 16)) = -1 then
                cmb.AddItem(IntToHex(pUInt64Tmp^, 16), nil);
              if cmb.Text = IntToHex(pUInt64Tmp^, 16) then
                copyPacketToGraphPTP(pDS, ptpDS, cmb.Text);
            end;
          end;
        pDS.Next;
      end;
    finally
      pDS.Bookmark := bm;
      pDS.EnableControls;
    end;
  finally
    ptpDS.EnableControls;
  end;
end;

procedure copyPacketsToGraphSensors(pDS, sDS: TDataSet; cmb: TComboBox);
var
  bm:         TBookmarkStr; // Deprecated infavor of TBookmark, but DataSet still uses it
  strTmp:     String;
  pUInt16Tmp: PWord;
  pUInt64Tmp: PQWord;
  nwkCmd:     Word;
begin
  sDS.DisableControls;
  try
    sDS.First;
    while not sDS.EOF do
      sDS.Delete;

    pDS.DisableControls;
    bm := pDS.Bookmark;
    try
      pDS.First;
      while not pDS.EOF do
      begin
        if not (pDS.FieldByName('Id').IsNull or pDS.FieldByName('Length').IsNull or
          pDS.FieldByName('Command').IsNull or pDS.FieldByName('Data').IsNull) then
          { test for sensor packet from network }
          if (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) or
            (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_RECEIVE_DATA_INDICATION) or
            (pDS.FieldByName('Command').AsInteger = szsCmd_LOCAL_SENSORS_REPORT) then
          begin
            strTmp := pDS.FieldByName('Data').AsString;
            { Verify if address of packet source is on the dropdown list }
            if (pDS.FieldByName('Command').AsInteger = ccaiCmd_ZB_SEND_DATA_REQUEST) then
            begin
              pUInt16Tmp := @strTmp[3];
              nwkCmd     := pUInt16Tmp^;
              pUInt64Tmp := @strTmp[1 + 8];
            end
            else if (pDS.FieldByName('Command').AsInteger =
              ccaiCmd_ZB_RECEIVE_DATA_INDICATION) then
            begin
              pUInt16Tmp := @strTmp[3];
              nwkCmd     := pUInt16Tmp^;
              pUInt64Tmp := @strTmp[1 + 6];
            end
            else
            begin
              nwkCmd     := nwkCmd_RPT;
              pUInt64Tmp := @strTmp[1];
            end;
            if nwkCmd = nwkCmd_RPT then
            begin
              if cmb.Items.IndexOf(IntToHex(pUInt64Tmp^, 16)) = -1 then
                cmb.AddItem(IntToHex(pUInt64Tmp^, 16), nil);
              if cmb.Text = IntToHex(pUInt64Tmp^, 16) then
                copyPacketToGraphSensors(pDS, sDS, cmb.Text);
            end;
          end;
        pDS.Next;
      end;
    finally
      pDS.Bookmark := bm;
      pds.EnableControls;
    end;
  finally
    sDS.EnableControls;
  end;
end;

procedure FillmdsCommands(DataSet: TDataSet);
begin
  DataSet.AppendRecord([ccaiCmd_ZB_START_REQUEST, 'CC2480 Start request']);
  DataSet.AppendRecord([ccaiCmd_ZB_BIND_DEVICE, 'CC2480 Bind device']);
  DataSet.AppendRecord([ccaiCmd_ZB_ALLOW_BIND, 'CC2480 Bind timeout']);
  DataSet.AppendRecord([ccaiCmd_ZB_SEND_DATA_REQUEST, 'CC2480 Send data request']);
  DataSet.AppendRecord([ccaiCmd_ZB_READ_CONFIGURATION, 'CC2480 Read configuration']);
  DataSet.AppendRecord([ccaiCmd_ZB_WRITE_CONFIGURATION, 'CC2480 Write configuration']);
  DataSet.AppendRecord([ccaiCmd_ZB_GET_DEVICE_INFO, 'CC2480 Get device information']);
  DataSet.AppendRecord([ccaiCmd_ZB_APP_REGISTER_REQUEST, 'CC2480 App register request']);

  DataSet.AppendRecord([ccaiCmd_SYS_RESET_IND, 'CC2480 System Reset Indication']);
  DataSet.AppendRecord([ccaiCmd_ZB_START_CONFIRM, 'CC2480 Start confirm']);
  DataSet.AppendRecord([ccaiCmd_ZB_BIND_CONFIRM, 'CC2480 Bind confirm']);
  DataSet.AppendRecord([ccaiCmd_ZB_ALLOW_BIND_CONFIRM, 'CC2480 Allow bind confirm']);
  DataSet.AppendRecord([ccaiCmd_ZB_SEND_DATA_CONFIRM, 'CC2480 Send data confirm']);
  DataSet.AppendRecord([ccaiCmd_ZB_RECEIVE_DATA_INDICATION,
    'CC2480 Receive data indication']);

  DataSet.AppendRecord([ccaiCmd_ZB_START_REQUEST_RSP, 'CC2480 Start request ACK']);
  DataSet.AppendRecord([ccaiCmd_ZB_ALLOW_BIND_RSP, 'CC2480 Allow bind ACK']);
  DataSet.AppendRecord([ccaiCmd_ZB_SEND_DATA_REQUEST_RSP, 'CC2480 Data request ACK']);
  DataSet.AppendRecord([ccaiCmd_ZB_READ_CONFIGURATION_RSP,
    'CC2480 Read configuration response']);
  DataSet.AppendRecord([ccaiCmd_ZB_WRITE_CONFIGURATION_RSP,
    'CC2480 Write configuration response']);
  DataSet.AppendRecord([ccaiCmd_ZB_GET_DEVICE_INFO_RSP,
    'CC2480 Get device information response']);
  DataSet.AppendRecord([ccaiCmd_ZB_APP_REGISTER_REQUEST_RSP,
    'CC2480 App register request response']);

  DataSet.AppendRecord([szsCmd_INITIALIZATION, '##### Device Initialization']);
  DataSet.AppendRecord([szsCmd_LOCAL_SENSORS_REPORT, 'Local sensor report']);
  DataSet.AppendRecord([szsCmd_TEXT_OUTPUT, 'Text output']);
  DataSet.AppendRecord([szsCmd_MEMORY_DUMP, 'Memory dump']);
end;

function getPTPClockAcuracySpecification(n: Byte): String;
begin
  case n of
    $20: Result        := 'Accuracy <= 25ns';
    $21: Result        := 'Accuracy <= 100ns';
    $22: Result        := 'Accuracy <= 250ns';
    $23: Result        := 'Accuracy <= 1us';
    $24: Result        := 'Accuracy <= 2.5us';
    $25: Result        := 'Accuracy <= 10us';
    $26: Result        := 'Accuracy <= 25us';
    $27: Result        := 'Accuracy <= 100us';
    $28: Result        := 'Accuracy <= 250us';
    $29: Result        := 'Accuracy <= 1ms';
    $2a: Result        := 'Accuracy <= 2.5ms';
    $2b: Result        := 'Accuracy <= 10ms';
    $2c: Result        := 'Accuracy <= 25ms';
    $2d: Result        := 'Accuracy <= 100ms';
    $2e: Result        := 'Accuracy <= 250ms';
    $2f: Result        := 'Accuracy <= 1s';
    $30: Result        := 'Accuracy <= 10s';
    $31: Result        := 'Accuracy > 10s';
    $32 .. $7f: Result := 'Reserved';
    $80 .. $fd: Result := 'Unused';
    $fe: Result        := 'Accuracy Unknown';
    else
      Result := 'Reserved';
  end;
end;

function getPTPMessageTypeName(n: Byte): String;
begin
  case n of
    $0: Result       := 'Sync';
    $1: Result       := 'Delay_Req';
    $2: Result       := 'Pdelay_Req';
    $3: Result       := 'Pdelay_Resp';
    $4 .. $7: Result := 'Reserved';
    $8: Result       := 'Follow_Up';
    $9: Result       := 'Delay_Resp';
    $a: Result       := 'Pdelay_Resp_Follow_Up';
    $b: Result       := 'Announce';
    $c: Result       := 'Signaling';
    $d: Result       := 'Management';
    $e .. $f: Result := 'Reserved';
    else
      Result := 'Unknown';
  end;
end;

function getPTPTimeSourceName(n: Byte): String;
begin
  case n of
    $10: Result        := 'Atomic Clock';
    $20: Result        := 'GPS';
    $30: Result        := 'Terrestrial Radio';
    $40: Result        := 'PTP';
    $50: Result        := 'NTP';
    $60: Result        := 'Hand Set';
    $90: Result        := 'Other';
    $a0: Result        := 'Internal Oscilator';
    $f0 .. $fe: Result := 'Unused';
    else
      Result := 'Reserved';
  end;
end;

{
Packet format:
        -----+--------+---------+-------+------------
Bytes  |  1  |    1   |    2    | 0-253 |      1     |
       +-----+--------+---------+-------+------------+
Desc   | SOF | Length | Command | Data  |    FCS     |
        -----+--------+---------+-------+------------
Offset |  0  |    1   |    2    |   4   | Length + 4 |
on
pbytRxBuf

refer to swra175a.pdf from Texas Instruments.

}
procedure processPacket(var buf: AnsiString; DataSet: TDataSet; keepCurrent: Boolean);
var
  pBuf:        PByte;
  L, i, cmd:   Word;
  bytFCS:      Byte;
  validPkg:    Boolean;
  nextSOF:     Integer;
  TS:          TDateTime;
  TSMs:        Word;
  TimeDiffNs:  Int64;
  RealTimeDiffNs: Int64;
  DTsS:        qword;
  DTsNs:       Cardinal;
  tickSize:    Cardinal;
  newTickSize: Cardinal;
  adjNs:       Int64;
  pbytTmp:     PByte;
  PUInt32Tmp:  PCardinal;
  bm:          TBookmarkStr; // Deprecated infavor of TBookmark, but DataSet still uses it
begin
  while buf <> '' do // Try to read all possible comands
  begin
    L := Length(buf);
    if L < 5 then // Minimum packet: SOF + Lenght + cmd0 + cmd1 + FCS
      Exit;
    pBuf := @buf[1]; // point to 1st char of the string
    { Check for Start Of Frame (SOF, char 254) }
    if pBuf[0] <> 254 then
      validPkg := False
    else
    begin
      { Check for complete packet }
      if pBuf[1] + 5 > L then
        Exit;
      { FCS check, it is an xor with all bytes except SOF and FCS }
      bytFCS := pBuf[1];
      for i := pBuf[1] + 3 downto 2 do
        bytFCS := bytFCS xor pBuf[i];
      if bytFCS <> pBuf[pBuf[1] + 4] then // Compare received FCS
        validPkg := False
      else
      begin
        { Proccess packet }

        TS   := Now;
        TSMs := MilliSecondOf(TS);

        validPkg := True;

        cmd := pBuf[2];
        cmd := ((cmd shl 8) or pBuf[3]);

        { Current time message }
        if (cmd = szsCmd_CURRENTTIME) then
        begin
          DTsS    := 0;
          pbytTmp := @DTsS;
          for i := 0 to 5 do // for each byte, copy value from source to
            pbytTmp[i] := pBuf[4 + i]; // dest (don't go beyond)
          PUInt32Tmp   := @pBuf[4 + 6];
          DTsNs        := PUInt32Tmp^;
          PUInt32Tmp   := @pBuf[4 + 10];
          tickSize     := PUInt32Tmp^;
          PUInt32Tmp   := nil;

          adjNs := 0;

          if prevDTsS <> 0 then
          begin
            TimeDiffNs     := ((DTsS - prevDTsS) * 1000000000) + DTsNs - prevDTsNs;
            RealTimeDiffNs := round((TS - prevTS) * 24 * 3600 * 1000000000);

            if (TimeDiffNs > 0) and (RealTimeDiffNs > 0) then
            begin
              newTickSize := round(tickSize * (Double(RealTimeDiffNs) / TimeDiffNs));
              sendTickSize(newTickSize);
            end;
            adjNs := DateTimeToUnix(TS) - DTsS;
            adjNs := adjNs + round(frac(TS * 24 * 3600) * 1000000000) - DTsNs;
            sendTimeAdjust(adjNs);

            if RealTimeDiffNs >= 60 then
            begin
              prevDTsS  := DTsS;
              prevDTsNs := DTsNs;
              prevTS    := TS;
            end;

            prevDTsS := prevDTsS + (adjNs div 1000000000);
            adjNs    := adjNs mod 1000000000;
            if (-adjNs > prevDTsNs) then
            begin
              prevDTsS  := prevDTsS - 1;
              prevDTsNs := prevDTsNs + (1000000000 + adjNs);
            end
            else if (adjNs + prevDTsNs) > 1000000000 then
            begin
              prevDTsS  := prevDTsS + 1;
              prevDTsNs := prevDTsNs + adjNs - 1000000000;
            end
            else
              prevDTsNs := prevDTsNs + adjNs;
          end
          else
          begin
            prevDTsS  := DTsS;
            prevDTsNs := DTsNs;
            prevTS    := ts;
          end;
        end;

        { Suspend updates to the controls }
        DataSet.DisableControls;
        if keepCurrent then
          bm := DataSet.Bookmark;
        try
          { check number of lines and don't let to go over 10000 data rows }
          while DataSet.RecordCount >= 10000 do
          begin
            DataSet.First;
            if keepCurrent and (DataSet.Bookmark = bm) then
            begin
              DataSet.Next;
              if DataSet.EOF then
                keepCurrent := False
              else
                bm          := DataSet.Bookmark;
              DataSet.First;
            end;
            DataSet.Delete;
          end;

          { record data }
          DataSet.Append;
          DataSet['TimeStampUnix'] := DateTimeToUnix(TS - (TSMs / MSecsPerDay));
          DataSet['TimeStampMs']   := TSMs;
          DataSet['Length']        := pBuf[1];
          DataSet['Command']       := cmd;
          DataSet.FieldByName('Data').AsString := Copy(buf, 5, pBuf[1]);
          DataSet.Post;
        finally
          if keepCurrent then
            DataSet.Bookmark := bm;
          { Re-activate updates to the controls }
          DataSet.EnableControls;
        end;
      end;
    end;
    if validPkg then // delete the processed command
      Delete(buf, 1, pBuf[1] + 5)
    else
    begin // if command is invalid, delete the string until next possible command
      Delete(buf, 1, 1);
      nextSOF := Pos(#254, buf);
      if nextSOF = 0 then // No other packet start, wipe it.
        buf := ''
      else
        Delete(buf, 1, nextSOF - 1);
    end;
  end;
end;

function showInvalidLenPacket(expectedLen, Len: Byte; strHexBuf: AnsiString;
  Memo: TMemo): Boolean;
begin
  if expectedLen = Len then
    Result := False
  else
  begin
    Memo.Append('Invalid length, expected ' + IntToStr(expectedLen) +
      ', current ' + IntToStr(Len));
    Memo.Append('Data: ' + strHexBuf);
    Result := True;
  end;
end;


{
        ------------
Bytes  |   Size * 2 |
       +------------+
Desc   | known      |
       | neighbors  |
        ------------
Offset |      0     |
on
pBuf

}
procedure showHeartBeat(pBuf: PByte; size: Byte; Memo: TMemo);
var
  i:          Byte;
  PUInt16Tmp: PWord;
begin
  Memo.Append('Heart beat');
  if size = 0 then
    Memo.Append('No known neighbors')
  else
    Memo.Append('Known neighbors:');
  i := 0;
  while i < size do
  begin
    PUInt16Tmp := @pBuf[i];
    Memo.Append('  0x' + IntToHex(PUInt16Tmp^, 4) + ' foreign index: ' +
      IntToStr(pBuf[i + 2]));
    i += 3;
  end;
end;

procedure showPacket(DataSet: TDataSet; Memo: TMemo);
var
  i:          Integer;
  pUInt16Tmp: PWord;
  pUInt64Tmp: PQWord;
  strTmp:     AnsiString;
  pBuf:       PByte;
  Len:        Byte;
  Cmd:        Word;
  cmdName:    String;
  buf, strHexBuf: AnsiString;
  nwkCmd:     Word;
  nwkLen:     Word;
begin
  if not DataSet.FieldByName('Id').IsNull then
  begin
    len       := DataSet.FieldByName('Length').AsInteger;
    cmd       := DataSet.FieldByName('Command').AsInteger;
    cmdName   := DataSet.FieldByName('cmdName').AsString;
    buf       := DataSet.FieldByName('Data').AsString;
    strHexBuf := DataSet.FieldByName('DataHex').AsString;
    pBuf      := @buf[1]; //point to first byte on string

    Memo.Text := '0x' + IntToHex(Cmd, 4) + ' - ' + cmdName;
    Memo.Append('Length: ' + IntToStr(len));
    Memo.Append('');

    case cmd of
      ccaiCmd_ZB_START_REQUEST: // ZB_START_REQUEST
        showInvalidLenPacket(0, Len, strHexBuf, Memo);
      ccaiCmd_ZB_BIND_DEVICE: // ZB_BIND_DEVICE
      { ------------------------------------------
        Bytes  |    1   |     2     |      8      |
               +--------+-----------+-------------+
        Desc   | Create | CommandId | Destination |
                --------+-----------+-------------
        Offset |    0   |     1     |      3      |
        on
        pBuf }
        if not showInvalidLenPacket(11, Len, strHexBuf, Memo) then
        begin
          if pBuf[0] = 0 then
            Memo.Append('Action: delete bind')
          else
            Memo.Append('Action: create bind');
          pUInt16Tmp := @pBuf[1];
          Memo.Append('CommandId: 0x' + IntToHex(pUInt16Tmp^, 4));
          pUInt64Tmp := @pBuf[3];
          Memo.Append('Destination: 0x' + IntToHex(pUInt64Tmp^, 16));
        end;
      ccaiCmd_ZB_ALLOW_BIND: // ZB_ALLOW_BIND
      { -------+---------
        Bytes  |    1    |
               +---------+
        Desc   | Timeout |
                ---------
        Offset |    0    |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          case pBuf[0] of
            0: Memo.Append('Disallow bind');
            1..254: Memo.Append('Allow bind for ' + IntToStr(pBuf[0]) + 's.');
            $FF: Memo.Append('Allow bind indefinitely.');
          end;
      ccaiCmd_ZB_SEND_DATA_REQUEST: // ZB_SEND_DATA_REQUEST
      { -------+-------------+-----------+--------+-----+--------+----
        Bytes  |      2      |     2     |    1   |  1  |    1   | ...
               +-------------+-----------+--------+-----+--------+----
        Desc   | Destination | CommandId | Handle | Ack | Radius | ...
                -------------+-----------+--------+-----+--------+----
        Offset |      0      |     2     |    4   |  5  |    6   | ...
        on
        pBuf

               ----+-----+------
        Bytes  ... |  1  | 0-84 |
               ----+-----+------+
        Desc   ... | len | data |
               ----+-----+------
        Offset ... |  7  |   8  |
        on
        pBuf }
      begin
        pUInt16Tmp := @pBuf[2];
        nwkCmd     := pUInt16Tmp^;

        pUInt16Tmp := @pBuf[0];
        Memo.Append('Destination: 0x' + IntToHex(pUInt16Tmp^, 4));
        Memo.Append('Cmd: 0x' + IntToHex(nwkCmd, 4));
        Memo.Append('Handle: 0x' + IntToHex(pBuf[4], 2));
        Memo.Append('Ack: 0x' + IntToHex(pBuf[5], 2));
        Memo.Append('Radius: ' + IntToStr(pBuf[6]));
        Memo.Append('Len: ' + IntToStr(pBuf[7]));
        Memo.Append('');

        if (nwkCmd = nwkCmd_HEARTBEAT) then // Data is a Heart Beat packet
          showHeartBeat(@pBuf[8], pBuf[7], Memo)
        else if (nwkCmd = nwkCmd_RPT) and ((pBuf[7] = 44) or (pBuf[7] = 76)) then
          // Data is a sensor report
          // Data length must be 44 (without time adjust info) or 78 (with)
          showSensorReport(@pBuf[8], pBuf[7], Memo)
        else if (nwkCmd = nwkCmd_PTP) and (pBuf[7] >= 34) then // Data is PTP packet
          showPTPReport(@pBuf[8], Memo)
        else // unknown packet sent to network
          Memo.Append('Unknown data: ' + Copy(strHexBuf, 8 * 3, 252));
      end;
      ccaiCmd_ZB_READ_CONFIGURATION: // ZB_READ_CONFIGURATION
      { -------+----------
        Bytes  |     1    |
               +----------+
        Desc   | ConfigId |
                ----------
        Offset |     0    |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          Memo.Append('configId: 0x' + IntToHex(pBuf[0], 2));

      ccaiCmd_ZB_WRITE_CONFIGURATION: // ZB_WRITE_CONFIGURATION
      { -------+----------+-----+-------
        Bytes  |     1    |  1  | 1-128 |
               +----------+-----+-------+
        Desc   | ConfigId | Len | Value |
                ----------+-----+-------
        Offset |     0    |  1  |   2   |
        on
        pBuf }
      begin
        Memo.Append('configId: 0x' + IntToHex(pBuf[0], 2));
        Memo.Append('Length: ' + IntToStr(pBuf[1]));
        Memo.Append('Value: ' + Copy(strHexBuf, 2 * 3, 252));
      end;
      ccaiCmd_ZB_GET_DEVICE_INFO: // ZB_GET_DEVICE_INFO
      { -------+-------
        Bytes  |   1   |
               +-------+
        Desc   | param |
                -------
        Offset |   0   |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          Memo.Append('param: 0x' + IntToHex(pBuf[0], 2));

      ccaiCmd_ZB_APP_REGISTER_REQUEST: // ZB_APP_REGISTER_REQUEST
      { -------+-------------+--------------+----------+----
        Bytes  |      1      |       2      |     2    | ...
               +-------------+--------------+----------+----
        Desc   | AppEndPoint | AppProfileID | DeviceId | ...
                -------------+--------------+----------+----
        Offset |      0      |       1      |     3    |
        on
        pBuf

               ----+---------------+--------+------------------+----
        Bytes  ... |       1       |    1   |         1        | ...
               ----+---------------+--------+------------------+----
        Desc   ... | DeviceVersion | Unused | InputCommandsNum | ...
               ----+---------------+--------+------------------+----
        Offset ... |       5       |    6   |         7        |
        on
        pBuf

               ----+-------------------+-------------------+----
        Bytes  ... |    2 * inCmdNum   |         1         | ...
               ----+-------------------+-------------------+----
        Desc   ... | InputCommandsList | OutputCommandsNum | ...
               ----+-------------------+-------------------+----
        Offset ... |          8        |  8 + 2 * inCmdNum |
        on
        pBuf

               ----+--------------------
        Bytes  ... |    2 * outCmdNum   |
               ----+--------------------+
        Desc   ... | OutputCommandsList |
               ----+--------------------
        Offset ... |  9 + 2 * inCmdNum  |
        on
        pBuf }
      begin
        Memo.Append('appEndPoint: ' + IntToStr(pBuf[0]));
        pUInt16Tmp := @pBuf[1];
        Memo.Append('appProfileID: 0x' + IntToHex(pUInt16Tmp^, 4));
        pUInt16Tmp := @pBuf[3];
        Memo.Append('deviceID: 0x' + IntToHex(pUInt16Tmp^, 4));
        Memo.Append('deviceVersion: ' + IntToStr(pBuf[5]));
        Memo.Append('Unused: 0x' + IntToHex(pBuf[6], 2));
        Memo.Append('InputCommandsNum: ' + IntToStr(pBuf[7]));
        strTmp := 'InputCommandsList:';
        for i := 0 to pBuf[7] - 1 do
        begin
          pUInt16Tmp := @pBuf[8 + (2 * i)];
          AppendStr(strTmp, ' 0x' + IntToHex(pUInt16Tmp^, 4));
        end;
        Memo.Append(strTmp);
        Memo.Append('OutputCommandsNum: ' + IntToStr(pBuf[8 + (2 * pBuf[7])]));
        strTmp := 'OutputCommandsList:';
        for i := 0 to pBuf[8 + (2 * pBuf[7])] - 1 do
        begin
          pUInt16Tmp := @pBuf[9 + (2 * pBuf[7]) + (2 * i)];
          AppendStr(strTmp, ' 0x' + IntToHex(pUInt16Tmp^, 4));
        end;
        Memo.Append(strTmp);
      end;
      ccaiCmd_SYS_RESET_IND: // CC2480 SYS_RESET_IND
      { -------+--------+--------------+-----------+----------+----
        Bytes  |    1   |       1      |     1     |     1    | ...
               +--------+--------------+-----------+----------+----
        Desc   | Reason | TransportRev | ProductId | MajorRel | ...
                --------+--------------+-----------+----------+----
        Offset |    0   |       1      |     2     |     3    |
        on
        pBuf
               ----+----------+-------
        Bytes  ... |     1    |   1   |
               ----+----------+-------+
        Desc   ... | MinorRel | HWRev |
               ----+----------+-------
        Offset ... |     4    |   5   |
        on
        pBuf }
        if not showInvalidLenPacket(6, Len, strHexBuf, Memo) then
        begin
          case pBuf[0] of
            0: // Power-up
              Memo.Append('Power-up');
            1: // External reset
              Memo.Append('External reset');
            2: // Watch dog reset
              Memo.Append('Watch dog reset');
          end;
          Memo.Append('TransportRev: ' + IntToStr(pBuf[1]));
          Memo.Append('ProductId: ' + IntToStr(pBuf[2]));
          Memo.Append('SW Release: ' + IntToStr(pBuf[3]) + '.' + IntToStr(pBuf[4]));
          Memo.Append('HW Revision: ' + IntToStr(pBuf[5]));
        end;

      ccaiCmd_ZB_START_CONFIRM: // CC2480 ZB_START_CONFIRM
      { -------+--------
        Bytes  |    1   |
               +--------+
        Desc   | Status |
                --------
        Offset |    0   |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          Memo.Append('Status: 0x' + IntToHex(pBuf[0], 2));
      ccaiCmd_ZB_BIND_CONFIRM: // CC2480 ZB_BIND_CONFIRM
      { -------+-----------+--------
        Bytes  |      2    |    1   |
               +-----------+--------+
        Desc   | CommandId | Status |
                -----------+--------
        Offset |      0    |    2   |
        on
        pBuf }
        if not showInvalidLenPacket(3, Len, strHexBuf, memo) then
        begin
          pUInt16Tmp := @pBuf[0];
          Memo.Append('Command Id: 0x' + IntToHex(pUInt16Tmp^, 4));
          Memo.Append('Status: 0x' + IntToHex(pBuf[2], 2));
          Memo.Append(strTmp);
        end;
      ccaiCmd_ZB_ALLOW_BIND_CONFIRM: // CC2480 ZB_ALLOW_BIND_CONFIRM
      { -------+--------
        Bytes  |    2   |
               +--------+
        Desc   | Source |
                --------
        Offset |    0   |
        on
        pBuf

      NOTE:
      Bug reported in http://e2e.ti.com/support/low_power_rf/f/158/t/226767.aspx
      returns 3 bytes, should be 2, 3rd byte is said to be "extraneous".
      }
        if not showInvalidLenPacket(3, Len, strHexBuf, Memo) then
        begin
          pUInt16Tmp := @pBuf[0];
          Memo.Append('Remote Address: 0x' + IntToHex(pUInt16Tmp^, 4));
          Memo.Append('Extra Byte: 0x' + IntToHex(pBuf[2], 2));
        end;
      ccaiCmd_ZB_SEND_DATA_CONFIRM: // CC2480 ZB_SEND_DATA_CONFIRM
      { -------+--------+--------
        Bytes  |    1   |    1   |
               +--------+--------+
        Desc   | Handle | status |
                --------+--------
        Offset |    0   |    1   |
        on
        pBuf }
        if not showInvalidLenPacket(2, Len, strHexBuf, Memo) then
        begin
          Memo.Append('Handle: 0x' + IntToHex(pBuf[0], 2));
          Memo.Append('Status: 0x' + IntToHex(pBuf[1], 2));
        end;
      ccaiCmd_ZB_RECEIVE_DATA_INDICATION: // CC2480 ZB_RECEIVE_DATA_INDICATION
      { -------+--------+---------+-----+------
        Bytes  |    2   |    2    |  2  | 0-84 |
               +--------+---------+-----+------+
        Desc   | source | command | len | data |
                --------+---------+-----+------
        Offset |    0   |    2    |  4  |   6  |
        on
        pBuf }
      begin
        pUInt16Tmp := @pBuf[2];
        nwkCmd     := pUInt16Tmp^;
        pUInt16Tmp := @pBuf[4];
        nwkLen     := pUInt16Tmp^;
        pUInt16Tmp := @pBuf[0];
        Memo.Append('Source: 0x' + IntToHex(pUInt16Tmp^, 4));
        Memo.Append('Cmd: 0x' + IntToHex(nwkCmd, 4));
        Memo.Append('Len: ' + IntToStr(nwkLen));
        Memo.Append('');
        if (nwkCmd = nwkCmd_HEARTBEAT) and (pBuf[7] = 0) then
          // Data is a sensor report
          showHeartBeat(@pBuf[6], nwkLen, Memo)
        else if (nwkCmd = nwkCmd_RPT) and ((nwkLen = 44) or (nwkLen = 76)) then
          // Data is a sensor report, data length must be 44 or 78
          showSensorReport(@pBuf[6], nwkLen, Memo)
        else if (nwkCmd = nwkCmd_PTP) and (nwkLen >= 34) then
          // Data is PTP packet, larger than 34 bytes (PTP header)
          showPTPReport(@pBuf[6], Memo)
        else // unknown packet from network
          Memo.Append('  Unknown data: ' + Copy(strHexBuf, 6 * 3, 252));
      end;
      ccaiCmd_ZB_START_REQUEST_RSP: // CC2480 ZB_START_REQUEST
        showInvalidLenPacket(0, Len, strHexBuf, Memo);
      ccaiCmd_ZB_ALLOW_BIND_RSP: // CC2480 ZB_ALLOW_BIND
        showInvalidLenPacket(0, Len, strHexBuf, Memo);
      ccaiCmd_ZB_SEND_DATA_REQUEST_RSP: // CC2480 ZB_SEND_DATA_REQUEST
        showInvalidLenPacket(0, Len, strHexBuf, Memo);
      ccaiCmd_ZB_WRITE_CONFIGURATION_RSP: // CC2480 ZB_WRITE_CONFIGURATION
      { -------+--------
        Bytes  |    1   |
               +--------+
        Desc   | Status |
                --------+
        Offset |    0   |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          Memo.Append('Status: 0x' + IntToHex(pBuf[0], 2));

      ccaiCmd_ZB_GET_DEVICE_INFO_RSP: // CC2480 ZB_GET_DEVICE_INFO
      { -------+-------+-------
        Bytes  |   1   |   8   |
               +-------+-------+
        Desc   | param | value |
                -------+-------
        Offset |   0   |   1   |
        on
        pBuf }
        if not showInvalidLenPacket(9, Len, strHexBuf, Memo) then
        begin
          Memo.Append('param: 0x' + IntToHex(pBuf[0], 2));
          pUInt64Tmp := @pBuf[1];
          Memo.Append('value: 0x' + IntToHex(pUInt64Tmp^, 16));
        end;
      ccaiCmd_ZB_APP_REGISTER_REQUEST_RSP: // CC2480 ZB_APP_REGISTER_REQUEST
      { -------+--------
        Bytes  |    1   |
               +--------+
        Desc   | Status |
                --------+
        Offset |    0   |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
          Memo.Append('Status: 0x' + IntToHex(pBuf[0], 2));
      szsCmd_INITIALIZATION: // Initialization
      { -------+-------
        Bytes  |   1   |
               +-------+
        Desc   | Value |
                -------
        Offset |   0   |
        on
        pBuf }
        if not showInvalidLenPacket(1, Len, strHexBuf, Memo) then
        begin
          Memo.Append('####################');
          case pBuf[0] of
            0: // Start
              Memo.Append('Normal initiatization.');
            1: // Reset by WTD+
              Memo.Append('Reset by watch dog.');
          end;
          Memo.Append('');
        end;

      szsCmd_LOCAL_SENSORS_REPORT: // Local Sensors report
      {
                        |  Temperature  |    Voltage    |   Luminance   |
                --------+-------+-------+-------+-------+-------+-------+----
        Bytes  |    8   |   2   |   10  |   2   |   10  |   2   |   10  | ...
               +--------+-------+-------+-------+-------+-------+-------+----
        Desc   | source | Value | Time  | Value | Time  | Value | Time  | ...
               |        |       | stamp |       | stamp |       | stamp |
                --------+-------+-------+-------+-------+-------+-------+----
        Offset |    0   |   8   |   10  |   20  |   22  |   32  |   34  |
        on
        pBuf
               | MeanPathDelay |  Last Clock Adjust  |
                ---------------+---------+-----------
        Bytes  |       8       |    8    |     10    |
               +-------+-------+---------+-----------+
        Desc   |     Value     |  Value  |   Time    |
               |               |         |   stamp   |
                -------+-------+---------+-----------
        Offset |       44      |    52   |     60    |
        on
        pBuf

      Time stamp: 6 bytes = seconds part, 4 bytes = nano-seconds part
      Data length must be 44 for SZS and 78 for SZSPTP
      }
        if (Len = 44) or (Len = 76) then
          showSensorReport(pBuf, Len, Memo)
        else
          showInvalidLenPacket(76, Len, strHexBuf, Memo);

      szsCmd_TEXT_OUTPUT: // Text to display
      { -------+-------
        Bytes  | 0-253 |
               +-------+
        Desc   | Text  |
                -------
        Offset |   0   |
        on
        pBuf

      Text output
      }
      begin
        strTmp := Copy(buf, 1, Len);
          {$ifdef MSWINDOWS}
        for i := Length(strTmp) - 1 downto 1 do
          if strTmp[i] = #10 then
            Insert(#13, strTmp, i);
          {$endif}
        Memo.Append(strTmp);
      end;
      szsCmd_MEMORY_DUMP: // Memory dump
      { -------+-----------------+------
        Bytes  |         1       | 0-64 |
               +-----------------+------+
        Desc   | Status & Length | Data |
                -----------------+------
        Offset |         0       |   1  |
        on
        pBuf

      Status is first 2 bits and Length is last 6 bits: ssllllll

      }
        if len >= 2 then
        begin
          Memo.Append('Memory dump');
          case pBuf[0] shr 6 of
            0: Memo.Append('  Status: OK');
            1: Memo.Append('  Status: Incomplete');
            else
              Memo.Append('  Status: Unknown');
          end;
          Memo.Append('  Data: ' + Copy(strHexBuf, 4, (Len - 1) * 3));
        end
        else
          showInvalidLenPacket(2, Len, strHexBuf, Memo);
      else
        Memo.Append(strHexBuf);
    end;
  end;
end;

procedure showPtpReport(pBuf: PByte; Memo: TMemo);
var
  i:          Integer;
  pbytTmp:    PByte;
  pUInt16Tmp: PWord;
  pUInt32Tmp: PLongWord;
  pInt64Tmp:  PInt64;
  pUInt64Tmp: PQWord;
  UInt48Tmp:  QWord;
  strTmp:     AnsiString;
begin
  UInt48Tmp := 0;
  memo.Append('PTP Message');
  memo.Append('  Version: ' + IntToStr(pBuf[1] and $0F) + '; Type: 0x' +
    IntToHex((pBuf[0]) and ($0F), 1) + ' (' + getPTPMessageTypeName(
    (pBuf[0]) and ($0F)) + ')');

  pUInt16Tmp := @pBuf[2]; // messageLength
  memo.Append('  Length: ' + IntToStr(pUInt16Tmp^) + '; domain: ' + IntToStr(pBuf[4]));
  pUInt16Tmp := @pBuf[6]; // flagField
  pInt64Tmp  := @pBuf[8]; // correctionField
  memo.Append(Format('  Flags: 0x%0:4.4x; correctionField: %1:12.1d -> %2:8.1nns',
    [pUInt16Tmp^, pInt64Tmp^, pInt64Tmp^ / 65536]));

  pUInt64Tmp := @pBuf[20]; // sourcePortIdentity clockIdentity
  pUInt16Tmp := @pBuf[28]; // sourcePortIdentity portNumber
  memo.Append('  sourceClock: 0x' + IntToHex(pUInt64Tmp^, 16) +
    '; sourcePort: ' + IntToStr(pUInt16Tmp^));
  pUInt16Tmp := @pBuf[30]; // sequenceId
  memo.Append('  sequence: ' + IntToStr(pUInt16Tmp^) + '; control: 0x' +
    IntToHex(pBuf[32], 2) + '; logMessageInterval: ' + IntToStr(pBuf[33]));
  case (pBuf[0]) and ($0F) of
    // $0: ; // Sync
    // $1: ; // Delay_Req
    $2:  // Pdelay_Req
    { -------+--------+-----------------+----------
      Bytes  |   34   |        10       |    10    |
             +--------+-----------------+----------+
      Desc   | Header | originTimestamp | reserved |
              --------+-----------------+----------
      Offset |    0   |        34       |    44    |
      on
      pBuf

    }
    begin
      memo.Append('  Pdelay_Req message:');
      pbytTmp := @UInt48Tmp;
      for i := 0 to 5 do
        pbytTmp[i] := pBuf[34 + i]; // originTimeStamp, s
      pUInt32Tmp   := @pBuf[40]; // originTimestamp, ns
      memo.Append(Format('    originTimestamp: %0:25.9ns',
        [UInt48Tmp + (pUInt32Tmp^ * 1e-9)]));
      strTmp := '    reserved:';
      for i := 44 to 53 do
        strTmp := strTmp + ' ' + IntToHex(pBuf[i], 2);
      Memo.Append(strTmp);
    end;
    $3: // Pdelay_Resp
    { -------+--------+--------------------------+------------------------
      Bytes  |   34   |            10            |           10           |
             +--------+--------------------------+------------------------+
      Desc   | Header | responseReceiptTimestamp | requestingPortIdentity |
              --------+--------------------------+------------------------
      Offset |    0   |            34            |           44           |
      on
      pBuf

    }
    begin
      memo.Append('  Pdelay_Resp message:');
      pbytTmp := @UInt48Tmp;
      for i := 0 to 5 do
        pbytTmp[i] := pBuf[34 + i]; // responseReceiptTimestamp, s
      pUInt32Tmp   := @pBuf[40]; // responseReceiptTimestamp, ns
      memo.Append(Format('    responseReceiptTimestamp: %0:25.9ns',
        [UInt48Tmp + (pUInt32Tmp^ * 1e-9)]));
      pUInt64Tmp := @pBuf[44];
      pUInt16Tmp := @pBuf[52];
      memo.Append(Format('    requesting Clock: 0x%0:.16h, port: %1:.1d',
        [pUInt64Tmp^, pUInt16Tmp^]));
    end;
    // $8: ; // Follow_Up
    // $9: ; // Delay_Resp
    $a: // Pdelay_Resp_Follow_Up
    { -------+--------+-------------------------+------------------------
      Bytes  |   34   |            10           |           10           |
             +--------+-------------------------+------------------------+
      Desc   | Header | responseOriginTimestamp | requestingPortIdentity |
              --------+-------------------------+------------------------
      Offset |    0   |            34           |           44           |
      on
      pBuf

    }
    begin
      memo.Append('  Pdelay_Resp_Follow_Up message:');
      pbytTmp := @UInt48Tmp;
      for i := 0 to 5 do
        pbytTmp[i] := pBuf[34 + i]; // responseOriginTimestamp, s
      pUInt32Tmp   := @pBuf[40]; // responseOriginTimestamp, ns
      memo.Append(Format('    responseOriginTimestamp: %0:25.9ns',
        [UInt48Tmp + (pUInt32Tmp^ * 1e-9)]));
      pUInt64Tmp := @pBuf[44];
      pUInt16Tmp := @pBuf[52];
      memo.Append(Format('    requesting Clock: 0x%0:.16h, port: %1:.1d',
        [pUInt64Tmp^, pUInt16Tmp^]));
    end;
    $b: // Announce
    { -------+--------+-----------------+------------------+----------+----
      Bytes  |   34   |        10       |         2        |     1    |
             +--------+-----------------+------------------+----------+----
      Desc   | Header | originTimestamp | currentUtcOffset | reserved | ...
              --------+-----------------+------------------+----------+----
      Offset |    0   |        34       |        44        |    46    |
      on
      pBuf

             ----+----------------------+-------------------------+----
      Bytes  ... |           1          |            4            | ...
             ----+----------------------+-------------------------+----
      Desc   ... | grandmasterPriority1 | grandmasterClockQuality | ...
             ----+----------------------+-------------------------+----
      Offset ... |          47          |            48           | ...
      on
      pBuf

             ----+----------------------+---------------------+--------------+----
      Bytes  ... |           1          |          8          |       2      | ...
             ----+----------------------+---------------------+--------------+----
      Desc   ... | grandmasterPriority2 | grandmasterIdentity | stepsRemoved | ...
             ----+----------------------+---------------------+--------------+----
      Offset ... |          52          |          53         |      61      | ...
      on
      pBuf

             ----+------------
      Bytes  ... |      1     |
             ----+------------+
      Desc   ... | timeSource |
             ----+------------
      Offset ... |     63     |
      on
      pBuf

    Text output
    }
    begin
      memo.Append('  Announce message');
      pbytTmp := @UInt48Tmp;
      for i := 0 to 5 do
        pbytTmp[i] := pBuf[34 + i]; // originTimeStamp, s
      pUInt32Tmp   := @pBuf[40]; // originTimeStamp, ns
      pUInt16Tmp   := @pBuf[44]; // currentUtcOffset
      memo.Append(Format(
        '    originTimestamp: %0:25.9ns; currentUtcOffset: %1:.2ds',
        [UInt48Tmp + (pUInt32Tmp^ * 1e-9), pUInt16Tmp^]));
      memo.Append('    grandMaster:');
      memo.Append('      Priority1: ' + IntToStr(pBuf[47]));
      memo.Append('      ClockQuality.clockClass: ' + IntToStr(pBuf[48]));
      memo.Append('      ClockQuality.clockAccuracy: ' +
        getPTPClockAcuracySpecification(pBuf[49]));
      pUInt16Tmp := @pBuf[50]; // grandmasterClockQuality.currentUtcOffset
      memo.Append('      ClockQuality.offsetScaledLogVariance: ' +
        IntToStr(pUInt16Tmp^));
      memo.Append('      Priority2: ' + IntToStr(pBuf[52]));
      pUInt64Tmp := @pBuf[53]; // grandmasterIdentity
      memo.Append('      Identity: 0x' + IntToHex(pUInt64Tmp^, 16));
      pUInt16Tmp := @pBuf[61]; // stepsRemoved
      memo.Append('      stepsRemoved: ' + IntToStr(pUInt16Tmp^));
      memo.Append('      timeSource: ' + getPTPTimeSourceName(pBuf[63]));
    end;
      // $c: ; // Signaling
      // $d: ; // Management
    else
    begin
      strTmp     := '  Data: ';
      pUInt16Tmp := @pBuf[2]; // messageLength
      for i := 34 to pUInt16Tmp^ - 1 do
        strTmp := strTmp + ' ' + IntToHex(pBuf[i], 2);
      Memo.Append(strTmp);
    end;
  end;
end;

{
                  |  Temperature  |    Voltage    |   Luminance   | ...
          --------+-------+-------+-------+-------+-------+-------+----
  Bytes  |    8   |   2   |   10  |   2   |   10  |   2   |   10  | ...
         +--------+-------+-------+-------+-------+-------+-------+----
  Desc   | source | Value | Time  | Value | Time  | Value | Time  | ...
         |        |       | stamp |       | stamp |       | stamp |
          --------+-------+-------+-------+-------+-------+-------+----
  Offset |    0   |   8   |   10  |   20  |   22  |   32  |   34  |
  on
  pBuf
         ... | Parent | MeanPathDelay |  Last Clock Adjust  | Tick |
         ----+--------+---------------+---------+-----------+------
  Bytes  ... |    2   |       8       |    8    |     10    |   4  |
         ----+--------+---------------+---------+-----------+------+
  Desc   ... | Short  |     Value     |  Value  |   Time    | Tick |
             | Addr   |               |         |   stamp   | Size |
         ----+--------+---------------+---------+-----------+------
  Offset ... |   44   |       46      |    54   |     62    |  72  |
  on
  pBuf

Time stamp: 6 bytes = seconds part, 4 bytes = nano-seconds part
Data length must be 44 for SZS and 76 for SZSPTP

}
procedure showSensorReport(pBuf: PByte; size: Byte; Memo: TMemo);
var
  v:          Word; // data value on a report
  vTsS:       QWord; // Time stamp of a value, seconds part
  vTsNs:      Cardinal; // Time stamp of a value, Nanoseconds part
  vu64:       QWord; // unsigned 64 bits data value
  v64:        Int64; // 64 bits data value
  pbytTmp:    PByte;
  PUInt16Tmp: PWord;
  PUInt32Tmp: PCardinal;
  pUInt64Tmp: PQWord;
  PInt64Tmp:  PInt64;
  i:          Byte;
begin
  v     := 0;
  vTsS  := 0;
  vTsNs := 0;

  // Address of the source of the packet
  pUInt64Tmp := @pBuf[0];
  Memo.Append('Source of the report: 0x' + IntToHex(pUInt64Tmp^, 16));

  // Temperature
  PUInt16Tmp := @pBuf[8];
  v          := PUInt16Tmp^;
  pbytTmp    := @vTsS; // point to dest variable,
  for i := 0 to 5 do // for each byte, copy value from source to
    pbytTmp[i] := pBuf[8 + 2 + i]; // dest (don't go beyond)
  PUInt32Tmp   := @pBuf[8 + 2 + 6];
  vTsNs        := PUInt32Tmp^;
  // °C = ((A10/1024)*1500mV)-986mV)*1/3.55mV; where A10 is the v variablte value
  // so: temp ~= 0.4126320423 * v - 277.746479
  Memo.Append(Format('  Temperature: %0:10.2f°C @ %1:25.9ns',
    [0.4126320423 * v - 277.746479, vTsS + (vTsNs * 1e-9)]));

  // Voltage
  PUInt16Tmp := @pBuf[20];
  v          := PUInt16Tmp^;
  pbytTmp    := @vTsS;
  for i := 0 to 5 do
    pbytTmp[i] := pBuf[20 + 2 + i];
  PUInt32Tmp   := @pBuf[20 + 2 + 6];
  vTsNs        := PUInt32Tmp^;
  // Measurement of AVcc/2 is in range 0 to 1023 indicating voltage from 0 to 1.5V
  // so: volt := 1.5 * 2 * v / 1023 = v / 341
  Memo.Append(Format('  Voltage: %0:15.2fV @ %1:25.9ns',
    [v / 341, vTsS + (vTsNs * 1e-9)]));

  // Luminance
  PUInt16Tmp := @pBuf[32];
  v          := PUInt16Tmp^;
  pbytTmp    := @vTsS;
  for i := 0 to 5 do
    pbytTmp[i] := pBuf[32 + 2 + i];
  PUInt32Tmp   := @pBuf[32 + 2 + 6];
  vTsNs        := PUInt32Tmp^;
  // Not enough data or imprecise data from the sensor datasheet to use a unit for luminance
  Memo.Append(Format('  Luminance: %0:5.1d(unknown) @ %1:25.9ns',
    [v, vTsS + (vTsNs * 1e-9)]));

  if size = 76 then
  begin
    PUInt16Tmp := @pBuf[44];
    Memo.Append(Format('  Master: 0x%0:4.4x', [PUInt16Tmp^]));
    pUInt64Tmp := @pBuf[44 + 2];
    vu64       := pUInt64Tmp^;
    Memo.Append(Format('  MeanPathDelay from/to master: %0:24.0nns', [1.0 * vu64]));
    PInt64Tmp := @pBuf[44 + 10];
    v64       := PInt64Tmp^;
    pbytTmp   := @vTsS;
    for i := 0 to 5 do
      pbytTmp[i] := pBuf[44 + 18 + i];
    PUInt32Tmp   := @pBuf[44 + 18 + 6];
    vTsNs        := PUInt32Tmp^;
    Memo.Append(Format('  Last adjust: %0:10.0fns @ %1:25.9ns',
      [0.0 + v64, vTsS + (vTsNs * 1e-9)]));
    PUInt32Tmp := @pBuf[44 + 28];
    Memo.Append(Format('  Tick size: %0:7.0nns', [1.0 * PUInt32Tmp^]));
  end;
end;

procedure showRawData(header: String; var buf: AnsiString; Memo: TMemo);
var
  i:      Integer;
  strTmp: String;
begin
  strTmp := header;
  for i := 1 to Length(buf) do
    strTmp := strTmp + ' ' + IntToHex(Ord(buf[i]), 2);
  buf      := '';
  Memo.Append(strTmp);
end;

procedure sendCommand(cmd: ShortString; serial: TSdpoSerial);
var
  strTmp:    ShortString;
  bytFcs, i: Byte;
begin
  if Length(cmd) - 2 < 0 then
    raise Exception.Create('sendCommand: Invalid parameter');
  bytFcs := Length(cmd) - 2;
  for i := Length(cmd) downto 1 do
    bytFcs := bytFcs xor Ord(cmd[i]);
  strTmp   := #254 + chr(Length(cmd) - 2) + cmd + chr(bytFcs);
  serial.WriteData(strTmp);
end;

procedure sendTickSize(newTickSize: Cardinal);
var
  pUInt32Tmp: PCardinal;
  cmd:        ShortString;
begin
  cmd := #225#33;
  SetLength(cmd, 6);
  pUInt32Tmp  := @cmd[3];
  pUInt32Tmp^ := newTickSize;
  pUInt32Tmp  := nil;
  sendCommand(cmd, frmMain.Serial);
end;

procedure sendTimeAdjust(adjNs: Int64);
var
  pUInt64Tmp: PInt64;
  cmd:        ShortString;
begin
  cmd := #225#34;
  SetLength(cmd, 10);
  pUInt64Tmp  := @cmd[3];
  pUInt64Tmp^ := adjNs;
  pUInt64Tmp  := nil;
  sendCommand(cmd, frmMain.Serial);
end;


function strIif(test: Boolean; truePart, falsePart: String): String;
begin
  if test then
    Result := truePart
  else
    Result := falsePart;
end;

end.
